from __future__ import annotations
import os
from typing import Optional, Union
import uuid
from enum import Enum
from arcgis.auth.tools import LazyLoader
import re
import copy

arcgis = LazyLoader("arcgis")
Content = LazyLoader("arcgis.apps.storymap.story_content")
json = LazyLoader("json")
time = LazyLoader("time")


class Themes(Enum):
    """
    Represents the Supported Theme Type Enumerations.
    Example: story_map.theme(Theme.Slate)
    """

    SUMMIT = "summit"
    OBSIDIAN = "obsidian"
    RIDGELINE = "ridgeline"
    MESA = "mesa"
    TIDAL = "tidal"
    SLATE = "slate"


###############################################################################################################
class StoryMap(object):
    """
    A Story Map is a web map that has been thoughtfully created, given context, and provided
    with supporting information so it becomes a stand-alone resource. It integrates maps, legends,
    text, photos, and video and provides functionality, such as swipe, pop-ups, and time sliders,
    that helps users explore this content.

    ArcGIS StoryMaps is the next-generation storytelling tool in ArcGIS, and story authors are
    encouraged to use this tool to create stories. The Python API can help you create and edit
    your stories.

    Create a Story Map object to make edits to a story. Can be created from an item of type 'Story Map',
    an item id for that type of item, or if .nothing is passed, a new story is created from a generic draft.

    If an Item or item_id is passed in, only published changes or new drafts are taken from the Story Map.
    If you have a story with unpublished changes, they will not appear when you construct your story with the API.
    If you start to work on your Story that has unpublished changes and save from the Python API, your
    unpublished changes on the GUI will be overwritten with your work from the API.

    ===============     ====================================================================
    **Parameter**        **Description**
    ---------------     --------------------------------------------------------------------
    item                Optional String or Item. The string for an item id or an item of type
                        'Story Map'. If no item is passed, a new story is created and saved to
                        your active portal.
    ---------------     --------------------------------------------------------------------
    gis                 Optional instance of :class:`~arcgis.gis.GIS` . If none provided the active gis is used.
    ===============     ====================================================================
    """

    _properties = None
    _gis = None
    _itemid = None
    _item = None
    _resources = None

    def __init__(
        self,
        item: Optional[Union[arcgis.gis.Item, str]] = None,
        gis: Optional[arcgis.gis.GIS] = None,
    ):
        # Section: Set up gis
        if gis is None:
            # If no gis, find active env
            gis = arcgis.env.active_gis
            self._gis = gis
        else:
            self._gis = gis
        if gis is None or gis._portal.is_logged_in is False:
            # Check to see if user is authenticated
            raise Exception("Must be logged into a Portal Account")

        # Section: Set up existing story
        if item and isinstance(item, str):
            # Get item using the item id
            item = gis.content.get(item)
            if item is None:
                # Error with storymap in current gis
                raise ValueError(
                    "Cannot find storymap associated with this item id in your portal. Please check it is correct."
                )
        if item and isinstance(item, arcgis.gis.Item) and item.type == "StoryMap":
            # Set item properties from existing item
            self._item = item
            self._itemid = self._item.itemid
            self._resources = self._item.resources.list()
            # Create existing story
            self._create_existing_storymap()
        elif (
            item
            and isinstance(item, arcgis.gis.Item)
            and "StoryMap" not in item.keywords
        ):
            # Throw error if item is not of type Story Map
            raise ValueError("Item is not a Story Map")
        else:
            # If no item was provided create a new story map
            self._create_new_storymap()
        # Get the story url
        self._url = self._get_url()

    # ----------------------------------------------------------------------
    def _create_existing_storymap(self):
        # Get properties from most recent resource file.
        # Can have multiple drafts so need to account for this.
        # Draft file will be of form: draft_{13 digit timestamp}.json or draft.json
        saved_drafts = []
        for resource in self._resources:
            for key, val in resource.items():
                # Find all drafts in the resources and add to a list
                if key == "resource" and (
                    re.match("draft_[0-9]{13}.json", val) or re.match("draft.json", val)
                ):
                    saved_drafts.append(val)
        # Find the correct draft to use
        if len(saved_drafts) == 1:
            # Only one draft saved
            # Open JSON draft file for properties
            data = self._item.resources.get(saved_drafts[0], try_json=True)
            self._properties = data
        elif len(saved_drafts) > 1:
            # Multiple drafts saved
            # Remove draft.json because oldest one
            if "draft.json" in saved_drafts:
                idx = saved_drafts.index("draft.json")
                del saved_drafts[idx]
            # check remaining to find most recent
            start = saved_drafts[0][6:19]  # get only timestamp
            current = saved_drafts[0]
            for draft in saved_drafts:
                compare = draft[6:19]
                if start < compare:
                    start = compare
                    current = draft
            # Open most recent JSON draft file for properties
            data = self._item.resources.get(current, try_json=True)
            self._properties = data
        else:
            # Storymap has no draft json so look for published json
            data = self._item.resources.get("published_data.json", try_json=True)
            self._properties = data

    # ----------------------------------------------------------------------
    def _create_new_storymap(self):
        # Get template from _ref folder
        template = copy.deepcopy(arcgis.apps.storymap._ref.storymap_2)
        # Add correct by-line and locale
        template["nodes"]["n-aTn8ak"]["data"]["byline"] = self._gis._username
        template["nodes"]["n-4xkUEe"]["config"]["storyLocale"] = (
            self._gis.users.me.culture if self._gis.users.me.culture else "en-US"
        )

        # Create unique story node id
        story_node = "n-" + uuid.uuid4().hex[0:6]
        template["root"] = story_node
        template["nodes"][story_node] = template["nodes"]["n-4xkUEe"]
        del template["nodes"]["n-4xkUEe"]
        # Set properties for the story
        self._properties = template
        # Create text for resource call
        text = json.dumps(template)
        # Create a temporary title
        title = "StoryMap via Python %s" % uuid.uuid4().hex[:10]
        # Create draft resource name
        draft = "draft_" + str(int(time.time() * 1000)) + ".json"
        # Will be posted as a draft
        sm_version = self._gis._con.get("https://storymaps.arcgis.com/version")[
            "version"
        ]
        keywords = ",".join(
            [
                "arcgis-storymaps",
                "StoryMap",
                "Web Application",
                "smstatusdraft",
                "smversiondraft:" + sm_version,
                "python-api",
                "smeditorapp:python-api-" + arcgis.__version__,
                "smdraftresourceid:" + draft,
            ]
        )
        # Get default thumbnail for a new item
        thumbnail = self._get_thumbnail()
        # Set the item properties dict to add new item to active gis
        item_properties = {
            "title": title,
            "keywords": keywords,
            "type": "StoryMap",
        }
        # Add item to active gis and set properties
        item = self._gis.content.add(
            item_properties=item_properties, thumbnail=thumbnail
        )
        # Assign to story properties
        self._item = item
        self._itemid = item.itemid
        # Make a resource call with the template to create json draft needed
        self._add_resource(resource_name=draft, text=text, access="private")
        # Assign resources to item
        self._resources = self._item.resources.list()

    # ----------------------------------------------------------------------
    def _repr_html_(self):
        """
        HTML Representation for IPython Notebook
        """
        return self._item._repr_html_()

    # ----------------------------------------------------------------------
    def __str__(self):
        """Return the url of the storymap"""
        return self._url

    # ----------------------------------------------------------------------
    def __repr__(self):
        return self.__str__()

    # ----------------------------------------------------------------------
    def _refresh(self):
        """Load the latest data from the item"""
        if self._item:
            self._properties = json.loads(self._item.get_data())

    # ----------------------------------------------------------------------
    def _get_url(self) -> str:
        """
        Private method to determine what the story url is. This is used to publish
        and have the correct path set.
        """
        if self._gis._is_agol:
            # Online
            self._url = "https://storymaps.arcgis.com/stories/{storyid}".format(
                storyid=self._itemid
            )
        else:
            # Enterprise
            self._url = "https://{portal}/apps/storymaps/stories/{storyid}".format(
                portal=self._gis.url, storyid=self._itemid
            )
        return self._url

    # ----------------------------------------------------------------------
    def _get_thumbnail(self) -> str:
        """
        Private method to get the default thumbnail path dependent on whether the
        user is Online or on Enterprise.
        """
        if self._gis._is_agol:
            thumbnail = "https://storymaps.arcgis.com/static/images/item-default-thumbnails/item.jpg"
        else:
            thumbnail = (
                self._gis._url
                + "/apps/storymaps/static/images/item-default-thumbnails/item.jpg"
            )
        return thumbnail

    # ----------------------------------------------------------------------
    def show(self, width: Optional[int] = None, height: Optional[int] = None):
        """
        Show a preview of the story. The default is a width of 700 and height of 300.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        width               Optional integer. The desired width to show the preview.
        ---------------     --------------------------------------------------------------------
        height              Optional integer. The desired height to show the preview.
        ===============     ====================================================================

        :return:
            An Iframe display of the story map if possible, else the item url is returned to be
            clicked on.
        """
        try:
            if self._item:
                width = 700 if width is None else width
                height = 350 if height is None else height
                from IPython.display import IFrame

                return IFrame(
                    src=self._item.url,
                    width=width,
                    height=height,
                    params="title=" + self._item.title,
                )
        except:
            return self._item.url

    # ----------------------------------------------------------------------
    @property
    def cover_date(self):
        """
        Get/Set the date type shown on the story cover.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        date_type           Optional String. Set the desired date type for the story cover.

                            ``Values: "first-published" | "last-published" | "none"``
        ===============     ====================================================================

        """
        root = self._properties["root"]
        return self._properties["nodes"][root]["config"]["coverDate"]

    # ----------------------------------------------------------------------
    @cover_date.setter
    def cover_date(self, date_type):
        """
        See cover_date property doc
        """
        # cover date is found in story node (i.e. root node id)
        root = self._properties["root"]
        self._properties["nodes"][root]["config"]["coverDate"] = date_type
        return self.cover_date

    # ----------------------------------------------------------------------
    @property
    def story_locale(self):
        """
        Get/Set the locale and language of the story.

        If your story was created with the Python API then the default is "en-US"
        """
        # story_locale is found in story node (i.e. root node id)
        root = self._properties["root"]
        return self._properties["nodes"][root]["config"]["storyLocale"]

    # ----------------------------------------------------------------------
    @story_locale.setter
    def story_locale(self, locale):
        """
        See story_locale property above
        """
        # cover date is found in story node (i.e. root node id)
        root = self._properties["root"]
        self._properties["nodes"][root]["config"]["storyLocale"] = locale
        return self.story_locale

    # ----------------------------------------------------------------------
    @property
    def properties(self):
        """This property returns the storymap's JSON."""
        return self._properties

    # ----------------------------------------------------------------------
    @property
    def nodes(self):
        """
        Get main nodes in order of appearance in the story. This will return a list
        of dictionaries specifying the node ids and the class content they correspond to.
        If there is no class for the content, a string is returned with the content type.
        """
        # node_dict contains key-value pairs where the value is the class instance
        node_dict = self._create_node_dict()
        # make the value the string representation of the class
        nodes = []
        for node in node_dict:
            nodes.append({k: node[k] for k in node})
        return nodes

    # ----------------------------------------------------------------------
    @property
    def content_list(self):
        """
        Get a list of all the content instances in order of appearance in the story.
        This returns a list of class instances for the content in the story.
        """
        contents = []
        # get the values from the nodes list and return only these
        nodes = self._create_node_dict()
        for node in nodes:
            content = list(node.values())[0]
            contents.append(content)
        return contents

    # ----------------------------------------------------------------------
    def _create_node_dict(self):
        """
        Method called by the nodes property and the get method. However, the nodes
        property will transform the keys whereas the get method needs they keys
        to be class instances.
        """
        # get rood node id since it is story node id
        root_id = self._properties["root"]
        # get list of children from story node
        children = self._properties["nodes"][root_id]["children"]
        nodes = self._properties["nodes"]

        node_order = []
        # for each node assign correct class type to be accessed if needed by user
        for child in children:
            # get only the main nodes and not the subnodes to be returned
            if child in nodes:
                node = self._assign_node_class(child)
                node_order.append({child: node})
        return node_order

    # ----------------------------------------------------------------------
    @property
    def actions(self):
        """
        Get list of action nodes.
        """
        actions = []
        if "actions" in self._properties:
            for action in self._properties["actions"]:
                node = self._assign_node_class(action["origin"])
                actions.append(node)
        return actions

    # ----------------------------------------------------------------------
    @property
    def navigation_list(self):
        """
        Get a list of the nodes that are linked in the navigation.
        """
        # navigation item has list of links corresponding to the text nodes in the navigation
        nav = self.get(type="navigation")[0]
        for key, value in nav.items():
            node_id = key
        try:
            links = self._properties["nodes"][node_id]["data"]["links"]
            node_ids = []
            for link in links:
                for key, value in link.items():
                    # Only return list of node_ids this way easier for navigation method
                    if key == "nodeId":
                        node_ids.append(value)
            return node_ids
        except:
            return None

    # ----------------------------------------------------------------------
    def get(self, node: Optional[str] = None, type: Optional[str] = None):
        """
        Get node(s) by type or by their id. Using this function will help grab a specific node
        from the story if a node id is provided. Set this to a variable and this way edits can be
        made on the node in the story.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        node                Optional string. The node id for the node that should be returned.
                            This will return the class of the node if of type story content.
        ---------------     --------------------------------------------------------------------
        type                Optional string. The type of nodes that user wants returned.
                            If none specified, list of all nodes returned.


                            Values: `image` | `video` | `audio` | `embed` | `webmap` | `text` |
                            `button` | `separator` | `expressmap` | `webscene` | `immersive`
        ===============     ====================================================================

        :return:
            If type specified: List of node ids and their types in order of appearance in the story map.

            If node_id specified: The node itself.


        .. code-block:: python

            >>> story = StoryMap(<story item>)

            # Example get by type
            >>> story.get(type = "text")
            Returns a list of all nodes of type text

            # Example by id
            >>> text = story.get(node= "<id for text node>")
            >>> text.properties
            Returns a specific node of type text

        """
        spec_type = []
        node_id = node
        if node_id and node_id not in self.properties["nodes"]:
            raise ValueError(
                "This node value is not in the story. "
                + "Please check that you have entered the correct node id. "
                + "To see all main nodes and their ids use the nodes property."
            )
        if type is None and node_id is None:
            # return all nodes in order
            return self.nodes
        elif node_id is not None:
            # check first if it's an action
            all_actions = self.actions
            for action in all_actions:
                id = list(action.keys())[0]
                if node_id == id:
                    return list(action.values())[0]
            # return a specific node
            all_nodes = self._create_node_dict()
            # find the node in the list and return it
            for node in all_nodes:
                id = list(node.keys())[0]
                if node_id == id:
                    return list(node.values())[0]
        else:
            # return all nodes of a certain type
            all_nodes = self._create_node_dict()
            for node in all_nodes:
                keyword = str(list(node.values())[0]).lower()
                if isinstance(keyword, str):
                    # Not a type of story content (i.e. navigation)
                    if type.lower() in keyword:
                        spec_type.append(node)
                else:
                    # Find all story content instances (i.e. Text)
                    # Map types are upercase and have spaces so handle
                    if type.lower() in keyword._type.lower().replace(" ", ""):
                        spec_type.append(node)
            return spec_type

    # ----------------------------------------------------------------------
    def cover(
        self,
        title: Optional[str] = None,
        type: str = None,
        summary: Optional[str] = None,
        by_line: Optional[str] = None,
        image: Optional[Content.Image] = None,
    ):
        """
        A story's cover is at the top of the story and always the first node.
        This method allows the cover to be edited by updating the title, byline, image, and more.
        Changing one part of the story cover will not change the rest of the story cover. If just the
        image is passed in then only the image will change.

        .. note::
            To change the date seen on the story cover, use the ``cover_date`` property.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        title               Optional string. The title of the StoryMap cover.
        ---------------     --------------------------------------------------------------------
        type                Optional string. The type of story cover to be used in the story.

                            ``Values: "full" | "sidebyside" | "minimal"``
        ---------------     --------------------------------------------------------------------
        summary             Optional string. The description of the story.
        ---------------     --------------------------------------------------------------------
        by_line             Optional string. Crediting the author(s).
        ---------------     --------------------------------------------------------------------
        image               Optional url or file path or :class:`~arcgis.apps.storymap.story_content.Image`
                            object. The cover image for the story cover.
        ===============     ====================================================================

        :return: Dictionary representation of the story cover node.

        .. code-block:: python

            story = StoryMap(<story item>)
            story.cover(title="My Story Title", type="minimal", summary="My little summary", by_line="python_dev")
            story.save()

        """
        # story cover is always first node
        dict_node = self.nodes[0]

        # get the node id
        for key, _ in dict_node.items():
            story_cover_node = key

        # get original data of story cover
        orig_data = self._properties["nodes"][story_cover_node]["data"]

        # set the new values, if any
        self._properties["nodes"][story_cover_node] = {
            "type": "storycover",
            "data": {
                "type": orig_data["type"] if type is None else type,
                "title": orig_data["title"] if title is None else title,
                "summary": orig_data["summary"] if summary is None else summary,
                "byline": orig_data["byline"] if by_line is None else by_line,
                "titlePanelPosition": orig_data["titlePanelPosition"]
                if by_line is None
                else "start",
            },
        }

        # set the cover image
        if image is not None:
            if not isinstance(image, Content.Image):
                image = Content.Image(image)
            if image.node not in self._properties["nodes"]:
                # must be added to story resources
                image._add_image(story=self)
            self._properties["nodes"][story_cover_node]["children"] = [image.node]
        else:
            # get original image
            if "children" in self._properties["nodes"][story_cover_node]:
                media = self._properties["nodes"][story_cover_node]["children"][0]
                self._properties["nodes"][story_cover_node]["children"] = [media]

        return self._properties["nodes"][story_cover_node]

    # ----------------------------------------------------------------------
    def navigation(
        self, nodes: Optional[list[str]] = None, hidden: Optional[bool] = None
    ):
        """
        Story navigation is a way for authors to add headings as
        links to allow readers to navigate between different sections
        of a story. The story navigation node takes ``TextStyle.HEADING`` text styles
        as its only allowed children.
        You can only have 30 :class:`~arcgis.apps.storymap.story_content.Text` child nodes
        as visible and act as links within a story.

        The text nodes must already exist in the story. Pass the list of node ids for the heading
        text nodes to assign them to the navigation.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        nodes               Optional list of nodes to include in the navigation. These nodes can
                            only be of style heading ("h2").
                            Include in order. This will override current list and order.

                            To see current list use ``navigation_list`` property.
        ---------------     --------------------------------------------------------------------
        hidden              Optional boolean. If True, the navigation is hidden.
        ===============     ====================================================================

        :return:
            List of nodes in the navigation.

        .. code-block:: python

            #Example
            >>> story = StoryMap("<existing story id>")
            >>> story.navigation_list

            >>> story.navigation(["<header node id>", "<header node id>"], False)
        """

        # Check if navigation node already exists
        for node, node_info in self._properties["nodes"].items():
            for key, val in node_info.items():
                if key == "type" and val == "navigation":
                    node_id = node

        links = []
        # If none is provided, set to what is already there
        if nodes is not None:
            # check nodes are correct and add in order with linkType
            for node in nodes:
                if self._properties["nodes"][node]["data"]["type"] == "h2":
                    links.append({"nodeId": node, "linkType": "story-heading"})
                elif self._properties["nodes"][node]["data"]["type"] == "h4":
                    links.append({"nodeId": node, "linkType": "credits-heading"})
        else:
            links = self._properties["nodes"][node_id]["data"]["links"]
        if hidden is None:
            hidden = self._properties["nodes"][node_id]["config"]["isHidden"]

        # Update navigation
        self._properties["nodes"][node_id] = {
            "type": "navigation",
            "data": {"links": links},
            "config": {"isHidden": hidden},
        }

        return self.navigation_list

    # ----------------------------------------------------------------------
    def theme(self, theme: Union[Themes, str] = Themes.SUMMIT):
        """
        Each story has a theme node in its resources. This method can be used to change the theme.
        To add a custom theme to your story, pass in the item_id for the item of type Story Map Theme.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        theme               Required Themes Style or custom theme item id.
                            The theme to set the story to.

                            Values: `SUMMIT` | `TIDAL` | `MESA` | `RIDGELINE` | `SLATE` | `OBSIDIAN` | `<item_id>`
        ===============     ====================================================================

        .. code-block:: python

            >>> from arcgis.apps.storymap import Themes

            >>> story = StoryMap()
            >>> story.theme(Themes.TIDAL)
        """
        # find the node corresponding to the story theme in resources
        for node, node_info in self._properties["resources"].items():
            for key, val in node_info.items():
                if key == "type" and val == "story-theme":
                    if isinstance(theme, Themes):
                        # theme comes from Themes class
                        self._properties["resources"][node]["data"][
                            "themeId"
                        ] = theme.value
                    if isinstance(theme, str):
                        # theme is an item of type Story Theme
                        self._properties["resources"][node]["data"][
                            "themeItemId"
                        ] = theme

    # ----------------------------------------------------------------------
    def credits(
        self,
        content: Optional[str] = None,
        attribution: Optional[str] = None,
        heading: Optional[str] = None,
        description: Optional[str] = None,
    ):
        """
        Credits are found at the end of the story and thus are always the last node.

        To create a credit, add the text that should be shown on each side of the divider.
        Content represents the text seen on the left side and attribution is in line with content
        on the right side of the divider. (i.e. 'content' | 'attribution')

        Adding ``content`` and ``attribution`` will add a new line to the credits and will not change previous
        credits.

        Adding ``heading`` and ``description`` will change what is currently in place.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        content             Optional String. The content to be added. (Seen on the left side of
                            the credits.)

                            Make sure text has '<strong> </strong>' tags.
                            Adds to the existing credits.
        ---------------     --------------------------------------------------------------------
        attribution         Optional String. The attribution to be added. (Seen on right side of
                            the credits.)
                            Adds to the existing credits.
        ---------------     --------------------------------------------------------------------
        heading             Optional String. Replace current heading for credits.
        ---------------     --------------------------------------------------------------------
        description         Optional String. Replace current description for credits.
        ===============     ====================================================================

        :return:
            A list of strings that are the node ids for the text nodes that belong to credits.

        .. code-block:: python

            #Example
            >>> story = StoryMap()
            >>> story.credits("Python Dev" , "Python API Team", "Thank You", "A big thank you to those who contributed")
        """
        # Find credit node
        dict_node = self.get(type="credits")[0]
        # Get credit node id
        for key, value in dict_node.items():
            credits_node = key
        if "children" in self._properties["nodes"][credits_node]:
            children = self._properties["nodes"][credits_node]["children"]
        else:
            children = []

        nodes = []
        if content or attribution:
            # Create new content node
            node = "n-" + uuid.uuid4().hex[0:6]
            self._properties["nodes"][node] = {
                "type": "attribution",
                "data": {"content": content, "attribution": attribution},
            }
            nodes.append(node)

        # Create new heading and remove old one
        if heading:
            # Create new content node
            node = "n-" + uuid.uuid4().hex[0:6]
            self._properties["nodes"][node] = {
                "type": "text",
                "data": {"text": heading, "type": "h4"},
            }
            nodes.append(node)
            # Find and remove node corresponding to current heading
            for child in children:
                if (
                    self._properties["nodes"][child]["type"] == "text"
                    and self._properties["nodes"][child]["data"]["type"] == "h4"
                ):
                    del self._properties["nodes"][child]
                    self._properties["nodes"][credits_node]["children"].remove(child)

        if description:
            # Create new content node
            node = "n-" + uuid.uuid4().hex[0:6]
            self._properties["nodes"][node] = {
                "type": "text",
                "data": {"text": description, "type": "paragraph"},
            }
            nodes.append(node)
            # Find and remove node corresponding to current heading
            for child in children:
                if (
                    self._properties["nodes"][child]["type"] == "text"
                    and self._properties["nodes"][child]["data"]["type"] == "paragraph"
                ):
                    del self._properties["nodes"][child]
                    self._properties["nodes"][credits_node]["children"].remove(child)

        # Add to children of credits
        for node_id in nodes:
            self._properties["nodes"][credits_node]["children"].append(node_id)
        return self._properties["nodes"][credits_node]["children"]

    # ----------------------------------------------------------------------
    def add(
        self,
        content: Optional[
            Union[
                Content.Image,
                Content.Video,
                Content.Audio,
                Content.Embed,
                Content.Map,
                Content.Button,
                Content.Text,
                Content.Gallery,
                Content.Timeline,
                Content.Sidecar,
            ]
        ] = None,
        caption: Optional[str] = None,
        alt_text: Optional[str] = None,
        display: str = None,
        position: Optional[int] = None,
    ):
        """
        Use this method to add content to your StoryMap. Content can be of various class types and when
        you add this content you can specify a caption, alt_text, display style, and the position
        at which it will be in your story.
        Not passing in any content means a separator will be added.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        content             Optional content of type:
                            :class:`~arcgis.apps.storymap.story_content.Image`,
                            :class:`~arcgis.apps.storymap.story_content.Gallery`,
                            :class:`~arcgis.apps.storymap.story_content.Video`,
                            :class:`~arcgis.apps.storymap.story_content.Audio`,
                            :class:`~arcgis.apps.storymap.story_content.Embed`,
                            :class:`~arcgis.apps.storymap.story_content.Map`,
                            :class:`~arcgis.apps.storymap.story_content.Text`,
                            :class:`~arcgis.apps.storymap.story_content.Button`,
                            :class:`~arcgis.apps.storymap.story_content.Timeline`,
                            :class:`~arcgis.apps.storymap.story_content.Sidecar`
                            :class:`~arcgis.apps.storymap.story_content.Swipe`,
                            :class:`~arcgis.apps.storymap.story_content.Separator`


                            If none is provided, a separator is added.
        ---------------     --------------------------------------------------------------------
        caption             Optional String. Custom text to caption the webmap.
        ---------------     --------------------------------------------------------------------
        alt_text            Optional String. Custom text to be used for screen readers.
        ---------------     --------------------------------------------------------------------
        display             Optional String. How the item will be displayed in the story map.

                            For Image, Video, Audio, or Map object.
                            Values: "standard" | "wide" | "full" | "float"

                            For Gallery:
                            Values: "jigsaw" | "square-dynamic"

                            For Embed:
                            Values: "card" | "inline"

                            For Swipe:
                            Values: "small" | "medium" | "large"
        ---------------     --------------------------------------------------------------------
        position            Optional Integer. Indicates the position in which the content will be
                            added. To see all node positions use the ``node`` property.
        ===============     ====================================================================

        :return: A String depicting the node id for the content that was added.

        .. code-block:: python

            new_story = StoryMap()

            # Example with Image
            >>> image1 = Image("<image-path>.jpg/jpeg/png/gif ")
            >>> new_node = new_story.add(image1, position = 2)

            # Example with Map
            >>> my_map = Map(<item-id of type webmap>)
            >>> new_node = new_story.add(my_map, "A map caption", "A new map alt-text")

            # Example to add a Separator
            >>> new_node = new_story.add()

            >>> print(new_story.nodes)

        """
        if content and content.node in self._properties["nodes"]:
            content.node = "n-" + uuid.uuid4().hex[0:6]

        # Node id included in all content except separator so create node id for that
        node_id = content.node if content is not None else "n-" + uuid.uuid4().hex[0:6]

        # Find instance of content and call correct method
        if isinstance(content, Content.Image):
            content._add_image(caption, alt_text, display, self)
        elif isinstance(content, Content.Gallery):
            content._add_gallery(caption, alt_text, display, self)
        elif isinstance(content, Content.Video):
            content._add_video(caption, alt_text, display, self)
        elif isinstance(content, Content.Audio):
            content._add_audio(caption, alt_text, display, self)
        elif isinstance(content, Content.Map):
            content._add_map(caption, alt_text, display, story=self)
        elif isinstance(content, Content.Embed):
            content._add_link(caption, alt_text, display, self)
        elif isinstance(content, Content.Button):
            content._add_button(self)
        elif isinstance(content, Content.Text):
            content._add_text(self)
        elif isinstance(content, Content.Timeline):
            content._add_timeline(self)
        elif isinstance(content, Content.Sidecar):
            content._add_sidecar(self)
        elif isinstance(content, Content.Swipe):
            content._add_swipe(caption, alt_text, display, self)
        else:
            content = Content.Separator(story=self, node_id=node_id)
            content._add_separator(story=self)

        # Add to story children
        self._add_child(node_id=node_id, position=position)
        return node_id

    # ----------------------------------------------------------------------
    def move(
        self, node_id: str, position: Optional[int] = None, delete_current: bool = False
    ):
        """
        Move a node to another position. The node currently at that position will
        be moved down one space. The node at the current position can be deleted
        instead of moved if `delete_current` is set to True.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        node_id             Required String. The node id for the content that will be moved. Find a
                            list of node order by using the ``nodes`` property.
        ---------------     --------------------------------------------------------------------
        position            Optional Integer. Indicates the position in which the content will be
                            added. If no position is provided, the node will be placed at the end.
        ---------------     --------------------------------------------------------------------
        delete_current      Optional Boolean. If set to True, the node at the current position will
                            be deleted instead of moved down one space. Default is False.
        ===============     ====================================================================

        .. code-block:: python

            new_story = StoryMap()

            # Example with Image
            >>> image1 = Image("<image-path>.jpg/jpeg/png/gif")
            >>> image2 = Image("<image-path>.jpg/jpeg/png/gif")
            >>> new_node = new_story.add(image1, "my caption", "my alt-text", "float", 2)
            >>> new_story.add(image2)
            >>> new_story.move(new_node, 3, False)

        """
        # Get list of story children
        root_id = self._properties["root"]
        children = self._properties["nodes"][root_id]["children"]

        # Remove node id from list since it will be added again at another position
        self._properties["nodes"][root_id]["children"].remove(node_id)

        # If delete_current is True then remove the node currently at this position
        if delete_current:
            if position == 0 or position == len(children):
                raise Exception(
                    "First and last nodes are reserved for Story Cover and Credits"
                )
            self._properties["nodes"][root_id]["children"].pop(position)

        # Add node to new position
        self._add_child(node_id, position)

    # ----------------------------------------------------------------------
    def save(
        self,
        title: Optional[str] = None,
        tags: Optional[list] = None,
        access: str = None,
        publish: bool = False,
    ):
        """
        This method will save your Story Map to your active GIS. The story will be saved
        with unpublished changes unless `publish` parameter is specified to True.

        The title only needs to be specified if a change is wanted, otherwise exisiting title
        is used.

        .. warning::
            Publishing your story through the Python API means it will not go through the Story Map
            issue checker. It is recommended to publish through the Story Maps builder if you
            want your story to go through the issue checker.

        .. warning::
            Changes to the published story may not be visible for up to one hour. You can open
            the story in the story builder to force changes to appear immediately and perform
            other optimizations, such as updating the story's social/SEO metadata.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        title               Optional string. The title of the StoryMap.
        ---------------     --------------------------------------------------------------------
        tags                Optional string. The tags of the StoryMap.
        ---------------     --------------------------------------------------------------------
        access              Optional string. The access of the StoryMap. If none is specified, the
                            current access is kept. This is used when `publish` parameter is set
                            to True.

                            Values: `private` | `org` | `public`
        ---------------     --------------------------------------------------------------------
        publish             Optional boolean. If True, the story is saved and also published.
                            Default is false so story is saved with unpublished changes.
        ===============     ====================================================================


        :return: The Item that was saved to your active GIS.

        """
        # Remove old draft item
        for resource in self._resources:
            if re.match("draft_[0-9]{13}.json", resource["resource"]) or re.match(
                "draft.json", resource["resource"]
            ):
                self._remove_resource(file=resource["resource"])

        # Add meta settings and change push meta so title doesn't get overwritten on publish at any point.
        if title:
            root = self._properties["root"]
            if "metaSettings" not in self._properties["nodes"][root]["data"]:
                self._properties["nodes"][root]["data"]["metaSettings"] = {
                    "title": None
                }
            self._properties["nodes"][root]["data"]["metaSettings"]["title"] = title
            self._properties["nodes"][root]["config"][
                "shouldPushMetaToAGOItemDetails"
            ] = False

        # Add new draft with time in milliseconds
        draft = "draft_" + str(int(time.time() * 1000)) + ".json"
        json_str = json.dumps(self._properties, ensure_ascii=False)
        self._add_resource(resource_name=draft, text=json_str, access="private")
        # get the story map version from endpoint
        sm_version = self._gis._con.get("https://storymaps.arcgis.com/version")[
            "version"
        ]
        # Find type keywords to use based on whether to publish or not
        # PUBLISH MODE
        if publish is True:
            # Remove old publish item
            for resource in self._resources:
                if (
                    "publish_data" in resource["resource"]
                    or "published_data" in resource["resource"]
                    or "publish" in resource["resource"]
                ):
                    self._remove_resource(file=resource["resource"])
            # Add new publish
            self._add_resource(
                resource_name="published_data.json", text=json.dumps(self._properties)
            )
            # Set the keywords
            # Start by getting the existing keywords and remove what will be replaced
            keywords = self._item.typeKeywords
            if "smstatusunpublishedchanges" in keywords:
                # changing to publish after
                idx = keywords.index("smstatusunpublishedchanges")
                del keywords[idx]
            if "smstatusdraft" in keywords:
                idx = keywords.index("smstatusdraft")
                del keywords[idx]
            for keyword in keywords:
                # iterate through since only know part of keyword we want to remove
                if (
                    "smdraftresourceid"
                    or "smpublisheddate"
                    or "smstatusdraft"
                    or "smpublisherapp"
                ) in keyword:
                    keywords.remove(keyword)
            new_keywords = [
                "smstatuspublished",
                "smversiondraft:" + sm_version,
                "smversionpublished:" + sm_version,
                "python-api",
                "smpublisherapp:python-api-" + arcgis.__version__,
                "smdraftresourceid:" + draft,
                "smpublisheddate:" + str(int(time.time() * 1000)),
            ]
            # Setting the keywords in a set will remove duplicates
            p = {
                "typeKeywords": list(set(keywords + new_keywords)),
                "text": json.dumps(self._properties),
                "url": self._url,
            }
            if title:
                p["title"] = title
            if tags:
                p["tags"] = tags

            # Find and set access
            sharing = access if access is not None else self._item.access
            p["access"] = sharing

            # Update the item and invoke share to have correct access
            self._item.update(item_properties=p)

            if sharing == "private":
                self._item.share(everyone=False, org=False, groups=None)
            elif sharing == "org":
                self._item.share(org=True)
            elif sharing == "public":
                self._item.share(everyone=True)

            if (
                self._gis._con._session.auth
                and self._gis._con._session.auth.token is not None
            ):
                # Make a call to the StoryMaps publish endpoint
                self._gis._con.post(
                    path=self._url + "/publish",
                    params={"f": "json", "token": self._gis._con._session.auth.token},
                )
        else:
            # Set the type keywords
            keywords = self._item.typeKeywords
            previously_published = False
            for keyword in keywords:
                if "smpublisheddate" in keyword:
                    # Update the date in new keywords
                    previously_published = True
                    keywords.remove(keyword)
                elif (
                    "smstatuspublished" in keyword
                    or "smstatusdraft" in keyword
                    or "smdraftresourceid" in keyword
                    or "smeditorapp" in keyword
                    or "Copy Item" in keyword
                ):
                    # Remove old keywords and will be replaced in new keywords
                    keywords.remove(keyword)
            if previously_published is True:
                # Unpublished changes mode
                new_keywords = [
                    "smstatusunpublishedchanges",
                    "smversiondraft:" + sm_version,
                    "python-api",
                    "smeditorapp:python-api-" + arcgis.__version__,
                    "smdraftresourceid:" + draft,
                    "smversionpublished:" + sm_version,
                    "smpublisheddate:" + str(int(time.time() * 1000)),
                ]
            if previously_published is False:
                # Draft mode
                new_keywords = [
                    "smstatusdraft",
                    "smversiondraft:" + sm_version,
                    "python-api",
                    "smeditorapp:python-api-" + arcgis.__version__,
                    "smdraftresourceid:" + draft,
                ]
            # Pass through set first to remove duplicates
            p = {"typeKeywords": list(set(keywords + new_keywords))}
            if title:
                p["title"] = title
            if tags:
                p["tags"] = tags
            # access does not change when only saving
            p["access"] = self._item.access
            self._item.update(item_properties=p)

        self._item = self._gis.content.get(self._itemid)
        return self._item

    # ----------------------------------------------------------------------
    def delete_story(self):
        """
        Deletes the story item.
        """
        # Check if item id exists
        item = self._gis.content.get(self._itemid)
        return item.delete()

    # ----------------------------------------------------------------------
    def duplicate(self, title: Optional[str] = None):
        """
        Duplicate the story. All items will be duplicated as they are. This allows you to create
        a story template and duplicate it when you want to work with it.

        It is highly recommended that once the duplicate is created, open it in Story Maps
        builder to ensure the issue checker finds any issues before editing.

        .. note::
            Can be used with ArcGIS Online or with ArcGIS Enterprise starting 10.8.1

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        title               Optional string. The title of the duplicated story. Only available for
                            ArcGISOnline.
        ===============     ====================================================================

        :return:
            The Item that was created.

        .. code-block:: python

            # Example for ArcGIS Online
            >>> story = StoryMap(<story item>)
            >>> story.duplicate("A Story Copy")

            # Example for ArcGIS Enterprise
            >>> story = StoryMap(<story item>)
            >>> story.duplicate()
        """
        # get the item to copy
        item = self._gis.content.get(self._itemid)

        # enterprise copy_item starting at 10.8.1
        if item._portal.is_arcgisonline is False and self._gis.version < [8, 2]:
            clone = self._gis.content.clone_items(items=[item])
        else:
            clone = item.copy_item(
                title="(Copy) " + self._item.title if title is None else title,
                include_resources=True,
                include_private=True,
            )
        # save to update keywords
        clone_story = StoryMap(clone.id)
        return clone_story.save()

    # ----------------------------------------------------------------------
    def copy_content(self, target_story: StoryMap, node_list: list):
        """
        Copy the content from one story to another. This will copy the content
        indicated to the target story in the order they are provided. To change the
        order once the nodes are copied, use the `move()` method on the target story.

        .. note::
            Do not forget to save the target story once you are done copying and making
            any further edits.

        .. note::
            This method can take time depending on the number of resources. Each resource coming
            from a file must be copied over and heavy files, such as videos or audio, can be time
            consuming.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        target_story        Required StoryMap instance. The target story that the content will be
                            copied to.
        ---------------     --------------------------------------------------------------------
        node_list           Required list of strings. The list of node ids indicating the content
                            that will be copied to the target story.
        ===============     ====================================================================

        :return:
            True if all nodes have been successfully copied over.

        """
        # Step 1: Do Checks
        # Check that nodes exist in original story (children of source story contain all of node_list)
        story_children = self._properties["nodes"][self._properties["root"]]["children"]
        check = all(node in story_children for node in node_list)
        # Return an error if not all nodes are in the source story.
        if check is False:
            not_in_story = []
            for node in node_list:
                if node not in story_children:
                    not_in_story.append(node)
            raise ValueError(
                "These nodes are not in the story: "
                + str(not_in_story)
                + ". Please check that the correct node ids are provided."
            )

        # Step 2: Create dictionaries for copying

        # Create node dict of all nodes to add, resource dict, and complete node list
        # Depending on node type, need to take different route to find all children
        original_nodes = node_list
        complete_node_list = []
        complete_node_dict = {}
        complete_resource_dict = {}
        resource_files = {}
        has_children = True

        # internal method to add to correct places
        def _add_to_dicts(node_add, comp_list, comp_node_dict, comp_res_dict):
            # add to complete list of nodes
            comp_list.append(node_add)
            # get the dictionary
            node_dict = self._properties["nodes"][node_add]
            comp_node_dict[node_add] = node_dict

            # find the resource node to add associated with node
            if "data" in node_dict:
                # iterate through values of dict to find any resources
                for _, value in node_dict["data"].items():
                    if isinstance(value, list):
                        for im in value:
                            # express maps keep their images in a list
                            _add_to_resources(im, comp_res_dict)
                    else:
                        _add_to_resources(value, comp_res_dict)

        def _add_to_resources(value, comp_res_dict):
            if isinstance(value, str):
                # check if value is a resource
                if "r-" in value:
                    resource_node = value
                    # get the resource dict
                    resource_dict = self._properties["resources"][resource_node]
                    comp_res_dict[resource_node] = resource_dict
                    if "resourceId" in resource_dict["data"]:
                        # some nodes keep the resource under resourceId key
                        name = resource_dict["data"]["resourceId"]
                        # get the resource file to add to new story
                        resource_file = self._item.resources.get(name)
                        resource_files[name] = resource_file
                    elif "itemId" in resource_dict["data"]:
                        name = resource_dict["data"]["itemId"]
                        # express map keeps resource under itemId key
                        if name.endswith(".json"):
                            # need to add draft_ in front to be one-to-one with builder
                            name = "draft_" + resource_dict["data"]["itemId"]
                            # get the json file draft
                            resource_file = self._item.resources.get(name)
                            resource_files[name] = resource_file

        # Begin populating dicts and list, assume there are children to begin with.
        while has_children is True:
            # new list of nodes to check at next iteration
            new_nodes = []
            for node in node_list:
                # add node info for copying
                _add_to_dicts(
                    node, complete_node_list, complete_node_dict, complete_resource_dict
                )
                # check type of node to see if need to find children
                node_children = self._has_children(node)
                # populate new list with next nodes to add
                if node_children:
                    for child in node_children:
                        new_nodes.append(child)
            # if list is not empty, keep going
            if new_nodes:
                has_children = True
                node_list = new_nodes
            # once list is empty, all children have been accounted for
            else:
                has_children = False

        # Step 3: Make any changes before copying over
        # existing target story node ids
        target_story_nodes = list(target_story._properties["nodes"].keys())

        if any(node in target_story_nodes for node in complete_node_list):
            # find the node and change it everywhere
            for node in complete_node_list:
                if node in target_story_nodes:
                    new_node = "n-" + uuid.uuid4().hex[0:6]
                    # replace node with new node in all places
                    # in the list passed in, if present
                    original_nodes = [s.replace(node, new_node) for s in original_nodes]
                    # in the dictionary of all nodes to copy
                    for key, value in complete_node_dict.items():
                        if key == node:
                            # replace old node id with new node id in keys
                            complete_node_dict[new_node] = complete_node_dict.pop(key)
                        if "children" in value:
                            # replace old node id with new node id if child of another node
                            if node in value["children"]:
                                complete_node_dict[key]["children"] = [
                                    s.replace(node, new_node) for s in value["children"]
                                ]

        # Step 4: Copy nodes to target story
        for key, value in complete_node_dict.items():
            target_story._properties["nodes"][key] = value
        for key, value in complete_resource_dict.items():
            target_story._properties["resources"][key] = value
        for key, value in resource_files.items():
            try:
                target_story._add_resource(file=value, resource_name=key)
            except:
                # express map, image editor, other created files will be here
                text = json.dumps(value)
                target_story._add_resource(resource_name=key, text=text)

        # Step 5: Add the node list to the story children
        for main_node in original_nodes:
            target_story._add_child(main_node)
        return True

    # ----------------------------------------------------------------------
    def _has_children(self, node):
        """
        Check if node has children and return list of children else None.
        """
        node_class = self._assign_node_class(node)
        if (
            isinstance(node_class, Content.Sidecar)
            or isinstance(node_class, Content.Gallery)
            or isinstance(node_class, Content.Timeline)
        ):
            return self._properties["nodes"][node]["children"]
        elif isinstance(node_class, Content.Swipe):
            return list(self._properties["nodes"][node]["data"]["contents"].values())
        elif isinstance(node_class, Content.MapTour):
            mt = self.get(node)
            return mt._children
        elif isinstance(node_class, str):
            if (
                "immersive" in node_class.lower()
                or "credits" in node_class.lower()
                or "event" in node_class.lower()
                or "carousel" in node_class.lower()
            ):
                return (
                    self._properties["nodes"][node]["children"]
                    if "children" in self._properties["nodes"][node]
                    else None
                )
        else:
            return None

    # ----------------------------------------------------------------------
    def _delete(self, node_id):
        # Check if node is in story
        if node_id not in self._properties["nodes"]:
            return False

        # Get list of nodes in the story
        root_id = self._properties["root"]
        children = self._properties["nodes"][root_id]["children"]

        # Remove from children of story
        if node_id in children:
            self._properties["nodes"][root_id]["children"].remove(node_id)
        # Remove from nodes dictionary
        del self._properties["nodes"][node_id]
        # Remove node from any immersive nodes.
        # A node can belong to an immersive narrative panel or an immersive slide
        for node in self._properties["nodes"]:
            if (
                "immersive" in self._properties["nodes"][node]["type"]
                and "children" in self._properties["nodes"][node]
            ):
                for child in self._properties["nodes"][node]["children"]:
                    # iterate through children to see if node is part of it
                    if child == node_id:
                        self._properties["nodes"][node]["children"].remove(node_id)

        return True

    # ----------------------------------------------------------------------
    def _add_child(self, node_id, position=None):
        """
        A story node has children. Children is a list of item nodes that are in
        the story. The order of the list determines the order that the nodes
        appear in the story. First and last nodes are reserved for story_cover
        and credits. The second node is always navigation. If visible is not set
        to True is simply won't be seen but stays in position 2.
        """
        # Get list of children in story
        root_id = self._properties["root"]
        last = len(self._properties["nodes"][root_id]["children"]) - 1

        if position and position < last and position != 0 and position != 1:
            # If the position adheres to rules then add node
            self._properties["nodes"][root_id]["children"].insert(position, node_id)
        elif position and (position == 0 or position == 1):
            # First and second node reserved for story cover and navigation
            # Add as third node if user specified position 0 or 1
            self._properties["nodes"][root_id]["children"].insert(2, node_id)
        else:
            # Last node is reserved for credits so add before this if user wanted last position
            self._properties["nodes"][root_id]["children"].insert(last, node_id)

    # ----------------------------------------------------------------------
    def _add_resource(self, file=None, resource_name=None, text=None, access="inherit"):
        """
        See :class:`~arcgis.gis.ResourceManager`
        """
        resource_manager = arcgis.gis.ResourceManager(self._item, self._gis)
        is_present = False
        if file:
            for resource in self._resources:
                if resource["resource"] in file:
                    is_present = True
                    resp = True
        properties = {
            "editInfo": {
                "editor": self._gis._username,
                "modified": str(int(time.time() * 1000)),
                "id": uuid.uuid4().hex[0:21],
                "app": "python-api",
            }
        }

        # access is inherited from item upon add, except for json where always private
        if is_present is False:
            resp = resource_manager.add(
                file=file,
                file_name=resource_name,
                text=text,
                access=access,
                properties=properties,
            )

        self._resources = self._item.resources.list()
        return resp

    # ----------------------------------------------------------------------
    def _remove_resource(self, file=None):
        """
        See :class:`~arcgis.gis.ResourceManager`
        """
        try:
            resource_manager = arcgis.gis.ResourceManager(self._item, self._gis)
            resp = resource_manager.remove(file=file)
            self._resources = self._item.resources.list()
            return resp
        except:
            # Resource cannot be found. Should not throw error
            return True

    # ----------------------------------------------------------------------
    def _assign_node_class(self, node_id):
        # Find the node type to assign to correct class
        node_type = self._properties["nodes"][node_id]["type"]
        # Create an instance of this class using existing node properties
        if node_type == "separator":
            node = Content.Separator(story=self, node_id=node_id)
        elif node_type == "image":
            node = Content.Image(story=self, node_id=node_id)
        elif node_type == "video":
            node = Content.Video(story=self, node_id=node_id)
        elif node_type == "audio":
            node = Content.Audio(story=self, node_id=node_id)
        elif node_type == "embed":
            # embed has subtype: video or link
            subtype = self._properties["nodes"][node_id]["data"]["embedType"]
            if subtype == "video":
                node = Content.Video(story=self, node_id=node_id)
            else:
                node = Content.Embed(story=self, node_id=node_id)
        elif node_type == "webmap":
            node = Content.Map(story=self, node_id=node_id)
        elif node_type == "text":
            node = Content.Text(story=self, node_id=node_id)
        elif node_type == "button":
            node = Content.Button(story=self, node_id=node_id)
        elif node_type == "swipe":
            node = Content.Swipe(story=self, node_id=node_id)
        elif node_type == "gallery":
            node = Content.Gallery(story=self, node_id=node_id)
        elif node_type == "timeline":
            node = Content.Timeline(story=self, node_id=node_id)
        elif node_type == "tour":
            node = Content.MapTour(story=self, node_id=node_id)
        elif node_type == "immersive":
            # immersive has subtype sidecar (more to add later)
            subtype = self._properties["nodes"][node_id]["data"]["type"]
            if subtype == "sidecar":
                node = Content.Sidecar(story=self, node_id=node_id)
            else:
                node = subtype
        elif node_type == "action-button":
            node = Content.MapAction(story=self, node_id=node_id)
        else:
            # if not of type story content then just return name of type
            node = node_type.capitalize()
        return node



from __future__ import annotations
import platform
from requests.auth import AuthBase
from urllib.parse import parse_qs
from arcgis.auth._auth._schain import SupportMultiAuth

from arcgis.auth.tools._lazy import LazyLoader
from arcgis.auth.tools import parse_url

HAS_KERBEROS = False
try:
    requests_kerberos = LazyLoader("requests_kerberos", strict=True)
    HAS_KERBEROS = True
except:
    HAS_KERBEROS = False


requests = LazyLoader("requests")

from ._utils import _split_username, assemble_url

__all__ = ["EsriKerberosAuth", "EsriWindowsAuth"]

HAS_SSPI = False
HAS_GSSAPI = False
HAS_KERBEROS = False
WINDOWS = False

if platform.platform().lower().find("windows") > -1:
    WINDOWS = True
    try:
        from ._negotiate import EsriHttpNegotiateAuth

        HAS_SSPI = True
    except:
        HAS_SSPI = False

try:
    requests_gssapi = LazyLoader("requests_gssapi", strict=True)
    HAS_GSSAPI = True
except:
    HAS_GSSAPI = False

try:
    requests_kerberos = LazyLoader("requests_kerberos", strict=True)
    HAS_KERBEROS = True
except:
    HAS_KERBEROS = False


requests = LazyLoader("requests")


class EsriWindowsAuth(AuthBase, SupportMultiAuth):
    _token_url = None
    _server_log = None
    _tokens = None

    def __init__(
        self,
        username: str = None,
        password: str = None,
        referer: str = None,
        verify_cert: bool = True,
        **kwargs,
    ):
        self.legacy = kwargs.pop("legacy", False)
        self.proxies = kwargs.pop("proxies", None)
        self._server_log = {}
        self._tokens = {}
        self._token_url = None
        self.verify_cert = verify_cert
        if referer is None:
            self.referer = "http"
        else:
            self.referer = referer

        try:
            if not username and not password and HAS_SSPI:
                self.auth = EsriHttpNegotiateAuth()

            elif username and password and HAS_SSPI:
                self.auth = EsriHttpNegotiateAuth(username=username, password=password)
            elif WINDOWS == True and HAS_KERBEROS:
                uname_format = _split_username(username)
                prin = uname_format[0] + "@" + uname_format[1]
                self.auth = requests_kerberos.HTTPKerberosAuth(
                    principal=f"{prin}:{password}",
                )
            elif HAS_GSSAPI:
                if not username or not password:
                    self.auth = requests_gssapi.HTTPSPNEGOAuth()
                else:
                    try:
                        from ._ntlm import EsriHttpNtlmAuth

                        self.auth = EsriHttpNtlmAuth(
                            username=username, password=password
                        )

                    except Exception as ex:
                        raise ex

            else:
                raise ValueError(
                    "Could not login, please ensure pywin32>225 and pyspnego are installed."
                )
        except ImportError:
            raise Exception("NTLM authentication requires pyspnego module.")

    # ----------------------------------------------------------------------
    def __str__(self):
        return f"<{self.__class__.__name__}>"

    # ----------------------------------------------------------------------
    def __repr__(self):
        return f"<{self.__class__.__name__}>"

    def generate_portal_server_token(self, r, **kwargs):
        """generates a server token using Portal token"""
        parsed = parse_url(r.url)
        server_url = assemble_url(parsed)
        if (
            r.text.lower().find("invalid token") > -1
            or r.text.lower().find("token required") > -1
            or r.text.lower().find("token not found") > -1
            or r.status_code == 401
            or r.text.lower().find("Access to admin resources are not allowed".lower())
            > -1
        ) or server_url in self._server_log:
            expiration = 16000

            postdata = {
                "request": "getToken",
                "serverURL": server_url,
                "referer": self.referer or "http",
                "f": "json",
            }
            if expiration:
                postdata["expiration"] = expiration
            if server_url in self._server_log:
                token_url = self._server_log[server_url]
            else:
                info = requests.get(
                    server_url + "/rest/info?f=json",
                    auth=self.auth,
                    verify=self.verify_cert,
                    proxies=self.proxies,
                ).json()
                token_url = info["authInfo"]["tokenServicesUrl"]
                self._server_log[server_url] = token_url
            if server_url in self._tokens:
                token_str = self._tokens[server_url]
            else:
                token = requests.post(
                    token_url,
                    data=postdata,
                    auth=self.auth,
                    proxies=self.proxies,
                    verify=self.verify_cert,
                )
                token_str = token.json().get("token", None)
                if token_str is None:
                    return r
                self._tokens[server_url] = token_str
            # Recreate the request with the token
            #
            r.content
            r.raw.release_conn()
            r.request.headers["referer"] = self.referer or "http"

            if self.legacy and r.request.method == "GET":
                r.request.prepare_url(url=r.url, params={"token": token_str})
            elif self.legacy and r.request.method == "POST":
                data = parse_qs(r.request.body)
                data["token"] = token_str
                r.request.prepare_body(data, None, None)
            else:
                r.request.headers["X-Esri-Authorization"] = f"Bearer {token_str}"

            # r.request.headers["X-Esri-Authorization"] = f"Bearer {token_str}"
            _r = r.connection.send(r.request, **kwargs)
            _r.headers["referer"] = self.referer or "http"
            _r.headers["X-Esri-Authorization"] = f"Bearer {token_str}"
            _r.history.append(r)
            return _r
        return r

    # ----------------------------------------------------------------------
    @property
    def token(self) -> str:
        """
        Gets the token.  This is always `None` for `EsriWindowsAuth`

        :returns: String
        """
        return None

    # ----------------------------------------------------------------------
    def __call__(self, r):
        self.auth.__call__(r)
        r.register_hook("response", self.generate_portal_server_token)
        return r


###########################################################################
class EsriKerberosAuth(AuthBase, SupportMultiAuth):
    _token_url = None
    _server_log = None
    _tokens = None

    def __init__(
        self,
        referer: str | None = None,
        verify_cert: bool = True,
        *,
        username: str | None = None,
        password: str | None = None,
        **kwargs,
    ):
        """initializer"""
        if HAS_KERBEROS == False:
            raise ImportError(
                "requests_kerberos is required to use this authentication handler."
            )
        self.proxies = kwargs.pop("proxies", None)
        self.legacy = kwargs.pop("legacy", False)

        self._server_log = {}
        self._tokens = {}
        self._token_url = None
        self.verify_cert = verify_cert
        self._session: requests.Session = kwargs.pop("session", requests.Session())
        mutual_auth_lu = {
            1: requests_kerberos.REQUIRED,
            2: requests_kerberos.OPTIONAL,
            3: requests_kerberos.DISABLED,
        }
        mutual_auth = mutual_auth_lu[kwargs.pop("mutual_authentication", 3)]
        if referer is None:
            self.referer = "http"
        else:
            self.referer = referer

        try:
            if username and password:
                uname_format = _split_username(username)
                prin = uname_format[0] + "@" + uname_format[1]
                self.auth = requests_kerberos.HTTPKerberosAuth(
                    mutual_authentication=mutual_auth,
                    principal=f"{prin}:{password}",
                    **kwargs,
                )
            else:
                self.auth = requests_kerberos.HTTPKerberosAuth(
                    mutual_authentication=mutual_auth, **kwargs
                )
        except ImportError:
            raise Exception(
                "Kerberos authentication requires `requests_kerberos` module."
            )

    # ----------------------------------------------------------------------
    def __str__(self):
        return f"<{self.__class__.__name__}>"

    # ----------------------------------------------------------------------
    def __repr__(self):
        return f"<{self.__class__.__name__}>"

    # ----------------------------------------------------------------------
    def generate_portal_server_token(self, r, **kwargs):
        """generates a server token using Portal token"""
        parsed = parse_url(r.url)
        server_url = assemble_url(parsed)
        if (
            r.text.lower().find("invalid token") > -1
            or r.text.lower().find("token required") > -1
            or r.text.lower().find("token not found") > -1
            or r.text.lower().find("Access to admin resources are not allowed".lower())
            > -1
        ) or server_url in self._server_log:
            expiration = 16000

            postdata = {
                "request": "getToken",
                "serverURL": server_url,
                "referer": self.referer or "http",
                "f": "json",
            }
            if expiration:
                postdata["expiration"] = expiration
            if server_url in self._server_log:
                token_url = self._server_log[server_url]
            else:
                info = self._session.get(
                    server_url + "/rest/info?f=json",
                    auth=self.auth,
                    verify=self.verify_cert,
                    proxies=self.proxies,
                ).json()
                token_url = info["authInfo"]["tokenServicesUrl"]
                self._server_log[server_url] = token_url
            if server_url in self._tokens:
                token_str = self._tokens[server_url]
            else:
                token = self._session.post(
                    token_url,
                    data=postdata,
                    auth=self.auth,
                    verify=self.verify_cert,
                    proxies=self.proxies,
                )
                token_str = token.json().get("token", None)
                if token_str is None:
                    return r
                self._tokens[server_url] = token_str
            # Recreate the request with the token
            #
            r.content
            r.raw.release_conn()
            r.request.headers["referer"] = self.referer or "http"
            r.request.headers["X-Esri-Authorization"] = f"Bearer {token_str}"

            if self.legacy and r.request.method == "GET":
                r.request.prepare_url(url=r.url, params={"token": token_str})
            elif self.legacy and r.request.method == "POST":
                data = parse_qs(r.body)
                data["token"] = token_str
                r.request.prepare_body(data, None, None)
            else:
                r.request.headers["X-Esri-Authorization"] = f"Bearer {token_str}"

            _r = r.connection.send(r.request, **kwargs)
            _r.headers["referer"] = self.referer or "http"
            _r.headers["X-Esri-Authorization"] = f"Bearer {token_str}"
            _r.history.append(r)
            if _r.status_code == 401:
                _r2 = self.auth.authenticate_user(_r)
                return _r2
            return _r
        return r

    # ----------------------------------------------------------------------
    @property
    def token(self) -> str:
        """
        Gets the token.  This is always `None` for `KerberosAuth`

        :returns: String
        """
        return None

    # ----------------------------------------------------------------------
    def __call__(self, r):
        self.auth.__call__(r)
        r.register_hook("response", self.generate_portal_server_token)
        return r



from __future__ import annotations
from arcgis.gis import GIS
from typing import TypeVar

T = TypeVar("T")
V = TypeVar("V")


class InstancePreference:
    """
    Provides information about the available instances for notebook containers.

    ================  ===============================================================================
    **Parameter**      **Description**
    ----------------  -------------------------------------------------------------------------------
    url               Required String. The base url for the InstancePreference endpoints.
    ----------------  -------------------------------------------------------------------------------
    gis               Required GIS. The ArcGIS Online connection object.
    ================  ===============================================================================

    """

    def __init__(self, url: str, gis: GIS):
        """init"""
        self._url = url
        self._gis = gis

    @property
    def available(self) -> dict[T, V]:
        """
        Returns Information on the available notebook instances.

        :returns: dict[T,V]
        """
        url = f"{self._url}/availableInstanceTypes"
        params = {"f": "json"}
        return self._gis._con.get(url, params)

    @property
    def instances(self) -> dict[T, V]:
        """
        Returns a dictionary containing the available instance types for the system

        :return: dict[T,V]
        """
        url = f"{self._url}"
        params = {"f": "json"}
        return self._gis._con.get(url, params)



from __future__ import annotations
from arcgis._impl.common._mixins import PropertyMap
from arcgis._impl.common._isd import InsensitiveDict
from arcgis.apps.hub.pages import PageManager, Page
from datetime import datetime
from collections import OrderedDict
from urllib.parse import urlparse
import json
import os
import re
from arcgis.gis import Item
from functools import wraps


def _lazy_property(fn):
    """Decorator that makes a property lazy-evaluated."""
    # http://stevenloria.com/lazy-evaluated-properties-in-python/
    attr_name = "_lazy_" + fn.__name__

    @property
    @wraps(fn)
    def _lazy_property(self):
        if not hasattr(self, attr_name):
            setattr(self, attr_name, fn(self))
        return getattr(self, attr_name)

    return _lazy_property


class Site(OrderedDict):
    """
    Represents a site within a Hub. A site is a container for
    web accessible content.
    """

    def __init__(self, gis, site_item: Item):
        """
        Constructs an empty Site object
        """
        self.item = site_item
        self._gis = gis
        try:
            self._sitedict = self.item.get_data()
            self.definition = PropertyMap(self._sitedict)
        except:
            self.definition = None

    def __repr__(self):
        return '<%s title:"%s" owner:%s>' % (
            type(self).__name__,
            self.title,
            self.owner,
        )

    @property
    def itemid(self) -> str:
        """
        Returns the item id of the site item
        """
        return self.item.id

    @property
    def title(self) -> str:
        """
        Returns the title of the site item
        """
        return self.item.title

    @property
    def description(self) -> str:
        """
        Getter/Setter for the site description
        """
        return self.item.description

    @description.setter
    def description(self, value):
        self.item.description = value

    @property
    def owner(self) -> str:
        """
        Returns the owner of the site item
        """
        return self.item.owner

    @property
    def tags(self) -> str:
        """
        Returns the tags of the site item
        """
        return self.item.tags

    @tags.setter
    def tags(self, value):
        self.item.tags = value

    @property
    def url(self) -> str:
        """
        Returns the url of the site
        """
        return self.item.url

    @property
    def initiative_id(self):
        """
        Returns the initiative id (if available) of the site
        """
        try:
            return self.item.properties["parentInitiativeId"]
        except:
            return None

    @_lazy_property
    def initiative(self):
        """
        Returns the initiative object (if available) for the site
        """
        try:
            return self._gis.hub.initiatives.get(self.initiative_id)
        except:
            return None

    @property
    def content_group_id(self) -> str:
        """
        Returns the groupId for the content group
        """
        try:
            return self.item.properties["contentGroupId"]
        except:
            return self.initiative.content_group_id

    @property
    def collab_group_id(self) -> str:
        """
        Returns the groupId for the collaboration group
        """
        try:
            return self.item.properties["collaborationGroupId"]
        except:
            if self._gis.hub._hub_enabled:
                return self.initiative.collab_group_id
            else:
                return None

    @property
    def catalog_groups(self) -> str:
        """
        Return Site catalog groups
        """
        return self.definition["catalog"]["groups"]

    @property
    def layout(self) -> InsensitiveDict:
        """
        Return layout of a site
        """
        return InsensitiveDict(self.definition["values"]["layout"])

    @property
    def theme(self):
        """
        Return theme of a site
        """
        return InsensitiveDict(self.definition["values"]["theme"])

    @_lazy_property
    def pages(self) -> PageManager:
        """
        The resource manager for an Initiative's indicators.
        See :class:`~hub.sites.PageManager`.
        """
        return PageManager(self._gis, self)

    def add_content(self, items_list: list):
        """
        Adds a batch of items to the site content library.

        =====================     ====================================================================
        **Parameter**              **Description**
        ---------------------     --------------------------------------------------------------------
        items_list                Required list. A list of Item or item ids to add to the Site.
        =====================     ====================================================================
        """
        # If input list is of item_ids, generate a list of corresponding items
        if type(items_list[0]) == str:
            items = [self._gis.content.get(item_id) for item_id in items_list]
        else:
            items = items_list
        # Fetch existing sharing privileges for each item, to retain them after adding to content library
        for item in items:
            sharing = item.shared_with
            everyone = sharing["everyone"]
            org = sharing["org"]
            groups = sharing["groups"]
            # add current site's content group to list of groups to share to
            groups.append(self.content_group_id)
            # share item to this group
            status = item.share(everyone=everyone, org=org, groups=groups)
            if status["results"][0]["success"] == False:
                return status
        return status

    def add_catalog_group(self, group_id: str):
        """

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        group_id            Group id to be added to site catalog
        ===============     ====================================================================
        """
        if group_id not in self.catalog_groups:
            self.definition["catalog"]["groups"].append(group_id)
            self.item.update(item_properties={"text": self.definition})
            return self.catalog_groups

    def delete_catalog_group(self, group_id: str):
        """

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        group_id            Group id to be added to site catalog
        ===============     ====================================================================
        """
        if group_id not in self.catalog_groups:
            raise Exception(
                "Group is not a part of site catalog. Please check the group_id"
            )
        self.definition["catalog"]["groups"] = [
            group for group in self.catalog_groups if group != group_id
        ]
        self.item.update(item_properties={"text": self.definition})
        return self.catalog_groups

    def delete(self) -> bool:
        """
        Deletes the Enterprise site. If unable to delete, raises a RuntimeException.

        :return:
            A bool containing True (for success) or False (for failure).

        .. note::
            Deleting the `site` object (instead of the `Item` object) for the site is encouraged,
            to also delete the groups associated with the site.

        .. code-block:: python

            USAGE EXAMPLE: Delete a site successfully

            site1 = gis.sites.get('itemId12345')
            site1.delete()

            >> True
        """
        # Unlink site from pages. Delete page if not linked to other sites
        try:
            site_pages = self.pages.search()
            # If pages exist
            if len(site_pages) > 0:
                for page in site_pages:
                    # Unlink page (deletes if)
                    self.pages.unlink(page)
        # In case site definition is empty
        except:
            pass
        # Fetch Site Collaboration group if exists
        _collab_group = None
        try:
            _collab_group_id = self.collab_group_id
            _collab_group = self._gis.groups.get(_collab_group_id)
            _collab_group.protected = False
            _collab_group.delete()
        except:
            pass
        # Fetch Content Group
        _content_group_id = self.content_group_id
        _content_group = self._gis.groups.get(_content_group_id)
        # Disable delete protection on groups and site
        try:
            _content_group.protected = False
            self.item.protect(enable=False)
            # Delete groups, site
            _content_group.delete()
        except:
            pass
        if not self._gis._portal.is_arcgisonline:
            return self.item.delete()
        else:
            # Deleting hub sites
            if self.item is not None:
                # Fetch site data
                _site_data = self.definition
                # Disable delete protection on site
                self.item.protect(enable=False)
                # Fetch siteId from domain entry
                path = "https://hub.arcgis.com/api/v3/domains?siteId=" + self.itemid
                _site_domain = self._gis._con.get(path=path)
                _siteId = _site_domain[0]["id"]
                # Delete domain entry
                session = self._gis._con._session
                headers = {k: v for k, v in session.headers.items()}
                headers["Content-Type"] = "application/json"
                headers["Authorization"] = "X-Esri-Authorization"
                path = "https://hub.arcgis.com/api/v3/domains/" + _siteId
                _delete_domain = session.delete(url=path, headers=headers)
                if _delete_domain.status_code == 200:
                    # Delete site item
                    return self.item.delete()
                else:
                    return _delete_domain.content

    def reassign_to(self, target_owner: str):
        """
        Allows the administrator to reassign the Enterprise site object from one
        user to another.

        .. note::
            This will transfer ownership of all items (site, pages, content) and groups to the new target_owner.

        =====================     ====================================================================
        **Parameter**              **Description**
        ---------------------     --------------------------------------------------------------------
        target_owner              Required string. The new desired owner of the site.
        =====================     ====================================================================
        """
        # check if admin user is performing this action
        if "admin" not in self._gis.users.me.role:
            return Exception(
                "You do not have the administrator privileges to perform this action."
            )
        # check if core team is needed by checking the role of the target_owner
        if self._gis.users.get(target_owner).role == "org_admin":
            # check if the initiative comes with core team by checking owner's role
            if self._gis.users.get(self.owner).role == "org_admin":
                # fetch the core team for the initative
                core_team = self._gis.groups.get(self.collab_group_id)
                # fetch the contents shared with this team
                core_team_content = core_team.content()
                # check if target_owner is part of core team, else add them to core team
                members = core_team.get_members()
                if (
                    target_owner not in members["admins"]
                    or target_owner not in members["users"]
                ):
                    core_team.add_users(target_owner)
                # remove items from core team
                self._gis.content.unshare_items(core_team_content, groups=[core_team])
                # reassign to target_owner
                for item in core_team_content:
                    item.reassign_to(target_owner)
                # fetch the items again since they have been reassigned
                new_content_list = []
                for item in core_team_content:
                    item_temp = self._gis.content.get(item.id)
                    new_content_list.append(item_temp)
                # share item back to the content group
                self._gis.content.share_items(
                    new_content_list, groups=[core_team], allow_members_to_edit=True
                )
                # reassign core team to target owner
                core_team.reassign_to(target_owner)
            else:
                # create core team necessary for the initiative
                _collab_group_title = _collab_group_title + " Core Team"
                _collab_group_dict = {
                    "title": _collab_group_title,
                    "tags": [
                        "Hub Group",
                        "Hub Site Group",
                        "Hub Core Team Group",
                        "Hub Team Group",
                    ],
                    "access": "org",
                    "capabilities": "updateitemcontrol",
                    "membershipAccess": "org",
                    "snippet": "Members of this group can create, edit, and manage the site, pages, and other content related to hub-groups.",
                }
                collab_group = self._gis.groups.create_from_dict(_collab_group_dict)
                collab_group.protected = True
                self.collab_group_id = collab_group.id
        else:
            # just reassign the initiative and site items
            self.item.reassign_to(target_owner)
            site_pages = self.pages.search()
            # If pages exist
            if len(site_pages) > 0:
                for page in site_pages:
                    # Unlink page (deletes if)
                    page.item.reassign_to(target_owner)
        # fetch content group
        content_team = self._gis.groups.get(self.content_group_id)
        # reassign to target_owner
        content_team.reassign_to(target_owner)
        return self._gis.content.get(self.itemid)

    def search(self, query: str | None = None, item_type: list | None = None) -> list:
        """
        Search and filter content for a site.

        =====================     ====================================================================
        **Parameter**              **Description**
        ---------------------     --------------------------------------------------------------------
        query                     Optional string. Filters items by presence of search query in title.
        ---------------------     --------------------------------------------------------------------
        item_type                 Optional list. Returns items of particular type.
        =====================     ====================================================================

        :return:
           List of items shared with this site.

        .. code-block:: python

            USAGE EXAMPLE: Succcessfully fetch items of item_type 'Web Mapping Application'
            for particular query 'school' for site

            site1 = myHub.sites.get('itemId12345')
            site_apps = site1.search(query='school', item_type='Web Map')
            site_apps

            >> List of relevant items
        """
        groups = self.catalog_groups
        all_content = []
        for group_id in groups:
            group = self._gis.groups.get(group_id)
            try:
                all_content = all_content + group.content()
            except AttributeError:
                pass
        # eliminate duplicate items
        result = [i for n, i in enumerate(all_content) if i not in all_content[:n]]
        if query != None:
            result = [item for item in result if query.lower() in item.title.lower()]
        if item_type != None:
            result = [item for item in result if item.type == item_type]
        return result

    def update(self, site_properties=None, subdomain=None):
        """Updates the site.

        .. note::
            For site_properties, pass in arguments for only the properties you want to be updated.
            All other properties will be untouched.  For example, if you want to update only the
            site's description, then only provide the description argument in site_properties.

        =====================     ====================================================================
        **Parameter**              **Description**
        ---------------------     --------------------------------------------------------------------
        site_properties           Required dictionary. See URL below for the keys and values.
        ---------------------     --------------------------------------------------------------------
        subdomain                 Optional string. New subdomain for the site.
        =====================     ====================================================================

        To find the list of applicable options for argument site_properties -
        https://developers.arcgis.com/python/api-reference/arcgis.gis.toc.html#arcgis.gis.Item.update

        :return:
           A boolean indicating success (True) or failure (False).

        .. code-block:: python
            USAGE EXAMPLE: Update a site successfully

            site1 = myHub.sites.get('itemId12345')
            site1.update(site_properties={'description':'Description for site.'})

            >> True
        """
        _site_data = self.definition
        if site_properties:
            for key, value in site_properties.items():
                _site_data[key] = value
        if subdomain:
            # format subdomain if needed
            subdomain = subdomain.replace(" ", "-").lower()
            # Domain manipulation for new subdomain
            if self._gis._portal.is_arcgisonline:
                # Check for length of domain
                if len(subdomain + "-" + self._gis.properties["urlKey"]) > 63:
                    _num = 63 - len(self._gis.properties["urlKey"])
                    raise ValueError(
                        "Requested url too long. Please enter a subdomain shorter than %d characters"
                        % _num
                    )
                # Fetch siteId from domain entry
                path = "https://hub.arcgis.com/api/v3/domains?siteId=" + self.itemid
                _site_domain = self._gis._con.get(path=path)
                _siteId = _site_domain[0]["id"]
                client_key = _site_domain[0]["clientKey"]
                # Delete old domain entry
                session = self._gis._con._session
                headers = {k: v for k, v in session.headers.items()}
                headers["Content-Type"] = "application/json"
                headers["Authorization"] = "X-Esri-Authorization"
                path = "https://hub.arcgis.com/api/v3/domains/" + _siteId
                _delete_domain = session.delete(url=path, headers=headers)
                # if deletion is successful
                if _delete_domain.status_code == 200:
                    # Create new domain entry

                    # Create domain entry for new site
                    _body = {
                        "hostname": subdomain
                        + "-"
                        + self._gis.properties["urlKey"]
                        + ".hub.arcgis.com",
                        "siteId": self.item.id,
                        "siteTitle": self.title,
                        "orgId": self._gis.properties.id,
                        "orgKey": self._gis.properties["urlKey"],
                        "orgTitle": self._gis.properties["name"],
                        "sslOnly": True,
                    }
                    headers = {k: v for k, v in session.headers.items()}
                    headers["Content-Type"] = "application/json"
                    headers["Authorization"] = "X-Esri-Authorization"
                    _new_domain = session.post(
                        url="https://hub.arcgis.com/api/v3/domains",
                        data=json.dumps(_body),
                        headers=headers,
                    )
                    if _new_domain.status_code == 200:
                        # define new domain, hostname and client_key
                        hostname = (
                            subdomain
                            + "-"
                            + self._gis.properties["urlKey"]
                            + ".hub.arcgis.com"
                        )
                        domain = self._gis.url[:8] + hostname
                        _client_key = _new_domain.json()["clientKey"]
                        # update initiative item
                        if self._gis.hub._hub_enabled:
                            self.initiative.item.update(item_properties={"url": domain})
                        # update site item and data
                        data = self.definition
                        data["values"]["defaultHostname"] = hostname
                        data["values"]["subdomain"] = subdomain
                        data["values"]["internalUrl"] = hostname
                        data["values"]["clientId"] = _client_key
                        if self.item.update(
                            item_properties={"url": domain, "text": data}
                        ):
                            return domain
                    # if creating new domain entry fails
                    else:
                        return _new_domain.content
                # if deleting old domain entry fails
                else:
                    return _delete_domain.content
            # For enterprise sites
            else:
                # Check for length of domain
                if len(subdomain) > 63:
                    raise ValueError(
                        "Requested url too long. Please enter a name shorter than 63 characters"
                    )
                typeKeywords = self.item.typeKeywords
                typeKeywords = [
                    keyword for keyword in typeKeywords if "hubsubdomain" not in keyword
                ]
                typeKeywords.append("hubsubdomain|" + subdomain)
                # Domain manipulation
                hostname = self._gis.url[7:-5] + "/apps/sites/#/" + subdomain
                domain = "https://" + hostname
                data = self.definition
                data["values"]["defaultHostname"] = hostname
                data["values"]["subdomain"] = subdomain
                data["values"]["internalUrl"] = hostname
                if self.item.update(
                    item_properties={
                        "typeKeywords": typeKeywords,
                        "url": domain,
                        "text": data,
                    }
                ):
                    return domain
        return self.item.update(_site_data)

    def update_layout(self, layout: dict) -> bool:
        """Updates the layout of the site.

        .. note::
            This operation can only be performed by the owner of the site or by an org administrator.

        =====================     ====================================================================
        **Parameter**              **Description**
        ---------------------     --------------------------------------------------------------------
        layout                    Required dictionary. The new layout dictionary to update to the site.
        =====================     ====================================================================

        :return:
           A boolean indicating success (True) or failure (False).

        .. code-block:: python

            USAGE EXAMPLE: Update a site successfully

            site1 = myHub.sites.get('itemId12345')
            site_layout = site1.layout
            site_layout.sections[0].rows[0].cards.pop(0)
            site1.update_layout(layout = site_layout)

            >> True
        """
        # Deleting the draft file for this site, if exists
        resources = self.item.resources.list()
        for resource in resources:
            if "draft-" in resource["resource"]:
                self.item.resources.remove(file=resource["resource"])
        # Update the data of the site
        self.definition["values"]["layout"] = layout._json()
        return self.item.update(item_properties={"text": self.definition})

    def update_theme(self, theme):
        """Updates the theme of the site.

        .. note::
            This operation can only be performed by the owner of the site or by an org administrator.

        =====================     ====================================================================
        **Parameter**              **Description**
        ---------------------     --------------------------------------------------------------------
        theme                     Required dictionary. The new theme dictionary to update to the site.
        =====================     ====================================================================

        :return:
           A boolean indicating success (True) or failure (False).

        .. code-block:: python

            USAGE EXAMPLE: Update a site successfully

            site1 = myHub.sites.get('itemId12345')
            site_theme = site1.theme
            site_theme.body.background = '#ffffff'
            site1.update_theme(theme = site_theme)

            >> True
        """
        # Deleting the draft file for this site, if exists
        resources = self.item.resources.list()
        for resource in resources:
            if "draft-" in resource["resource"]:
                path = (
                    self._gis.url
                    + "/sharing/rest/content/items/"
                    + self.itemid
                    + "/resources/"
                    + resource["resource"]
                    + "?token="
                    + self._gis._con.token
                )
                self.item.resources.remove(file=path)
        # Update the data of the site
        self.definition["values"]["theme"] = theme._json()
        return self.item.update(item_properties={"text": self.definition})


class SiteManager(object):
    """
    Helper class for managing sites within a Hub. This class is not created by users directly.
    An instance of this class, called 'sites', is available as a property of the Hub object. Users
    call methods on this 'sites' object to manipulate (add, get, search, etc) sites.
    """

    def __init__(self, gis, initiative=None):
        # If accessed from the gis route
        try:
            gis._portal
            self._gis = gis
            # For Hub (ArcGIS Online) sites
            if self._gis._portal.is_arcgisonline:
                self._hub = self._gis.hub
        # If accessed from the hub route for cloning from Enterprise to Hub
        except:
            self._hub = gis
            self._gis = self._hub.gis
        self.initiative = initiative

    def _create_and_register_site(
        self, site, subdomain, site_data, content_group_id, collab_group_id
    ):
        """
        Registers site as an app and Creates a domain entry for new site.
        Updates data with necessary attributes for a new site.
        """

        basemap = {}

        if self._gis._portal.is_arcgisonline:
            # register site as an app
            # _app_dict = site.register(app_type="browser", redirect_uris=[site.url])
            # client_key = _app_dict["client_id"]

            # Check for length of domain
            if len(subdomain + "-" + self._gis.properties["urlKey"]) > 63:
                _num = 63 - len(self._gis.properties["urlKey"])
                raise ValueError(
                    "Requested url too long. Please enter a name shorter than %d characters"
                    % _num
                )

            session = self._gis._con._session
            # Create domain entry for new site
            _body = {
                "hostname": subdomain
                + "-"
                + self._gis.properties["urlKey"]
                + ".hub.arcgis.com",
                "siteId": site.id,
                "siteTitle": site.title,
                "orgId": self._gis.properties.id,
                "orgKey": self._gis.properties["urlKey"],
                "orgTitle": self._gis.properties["name"],
                "sslOnly": True,
            }
            headers = {k: v for k, v in session.headers.items()}
            headers["Content-Type"] = "application/json"
            headers["Authorization"] = "X-Esri-Authorization"
            _new_domain = session.post(
                url="https://hub.arcgis.com/api/v3/domains",
                data=json.dumps(_body),
                headers=headers,
            )
            if _new_domain.status_code == 200:
                _siteId = _new_domain.json()["id"]
                _client_key = _new_domain.json()["clientKey"]
            else:
                return _new_domain
        else:
            # Check for length of domain
            if len(subdomain) > 63:
                raise ValueError(
                    "Requested url too long. Please enter a name shorter than 63 characters"
                )

        # for group_id in group_ids:
        try:
            site_data["catalog"]["groups"].append(content_group_id)
        except:
            site_data["values"]["groups"].append(content_group_id)
            site_data["values"]["uiVersion"] = "2.3"
        if self._gis._portal.is_arcgisonline:
            site_data["values"]["theme"]["globalNav"] = {}
            try:
                site_data["values"]["theme"]["globalNav"] = self._gis.properties[
                    "portalProperties"
                ]["sharedTheme"]["header"]
            except KeyError:
                site_data["values"]["theme"]["globalNav"] = {
                    "background": "#fff",
                    "text": "#000000",
                }
        site_data["values"]["title"] = site.title
        site_data["values"]["layout"]["header"]["component"]["settings"][
            "title"
        ] = site.title
        site_data["values"]["collaborationGroupId"] = collab_group_id
        site_data["values"]["subdomain"] = subdomain
        site_data["values"]["defaultHostname"] = site.url
        site_data["values"]["updatedBy"] = self._gis.users.me.username
        if self._gis._portal.is_arcgisonline:
            site_data["values"]["siteId"] = _siteId
            site_data["values"]["clientId"] = _client_key
        else:
            site_data["values"]["clientId"] = "arcgisonline"
        # Add collaboration group to gallery card only if it exists in the usual spot
        try:
            site_data["values"]["layout"]["sections"][6]["rows"][1]["cards"][0][
                "component"
            ]["settings"]["selectedGroups"][0]["id"] = collab_group_id
        except:
            pass
        # Link follow button to current initiative only if it exists in the usual spot
        if self._gis._portal.is_arcgisonline and self._hub._hub_enabled:
            try:
                site_data["values"]["layout"]["sections"][8]["rows"][1]["cards"][0][
                    "component"
                ]["settings"]["initiativeId"] = self.initiative.itemid
            except:
                pass
        site_data["values"]["map"] = self._gis.properties["defaultBasemap"]
        site_data["values"]["defaultExtent"] = self._gis.properties["defaultExtent"]

        return site_data

    def add(self, title, subdomain=None):
        """
        Adds a new site.

        .. note::
            Unicode characters are not allowed in the title of the site.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        title               Required string.
        ---------------     --------------------------------------------------------------------
        subdomain           Optional string. Available ONLY with Enterprise Sites.
        ===============     ====================================================================

        :return:
           The site if successfully added, None if unsuccessful.

        .. code-block:: python

            USAGE EXAMPLE: Add an open data site in Hub successfully

            site1 = myHub.sites.add(title='My first site')
            site1.item

        .. code-block:: python

            USAGE EXAMPLE: Add an initiative site successfully

            initiative_site = initiative1.sites.add(title=title)
            site1.item
        """

        siteId = None
        # Checking if subdomain is provided
        if subdomain:
            # disallow for AGO
            if self._gis._portal.is_arcgisonline:
                raise Exception(
                    "The option to add sites with custom subdomain is only available with Enterprise Sites. Please add this site without custom subdomain."
                )
            # re-format if given for Enterprise sites
            else:
                subdomain = subdomain.replace(" ", "-").lower()
        # re-format title if subdomain not provided
        else:
            subdomain = title.replace(" ", "-").lower()

        # Check if initiative or site needs to be created for this gis
        if self._gis._portal.is_arcgisonline:
            if self._hub._hub_enabled:
                if self.initiative is None:
                    raise Exception(
                        "Sites are created as part of an Initiative for your Hub. Please add a new initiative to proceed."
                    )

        # For sites in ArcGIS Online
        if self._gis._portal.is_arcgisonline:
            # Set item details
            item_type = "Hub Site Application"
            typekeywords = "Hub, hubSite, hubSolution, JavaScript, Map, Mapping Site, Online Map, OpenData, Ready To Use, selfConfigured, Web Map, Registered App"
            tags = ["Hub Site"]
            description = "DO NOT DELETE OR MODIFY THIS ITEM. This item is managed by the ArcGIS Hub application. To make changes to this site, please visit https://hub.arcgis.com/admin/"

            # Domain manipulation
            domain = (
                self._gis.url[:8]
                + subdomain
                + "-"
                + self._gis.properties["urlKey"]
                + ".hub.arcgis.com"
            )
            _request_url = "https://hub.arcgis.com/utilities/domains/" + domain[8:]
            session = self._gis._con._session
            headers = {k: v for k, v in session.headers.items()}
            headers["Content-Type"] = "application/json"
            headers["Authorization"] = "X-Esri-Authorization"
            response = session.get(
                url=f"https://hub.arcgis.com/api/v3/domains/" + domain[8:],
                headers=headers,
            )
            # Check if domain doesn't exist
            if response.status_code == 404:
                pass
            else:
                # If exists check if counter needs updating and update it
                try:
                    count = int(subdomain[-1])
                    count = count + 1
                    subdomain = subdomain[:-1] + str(count)
                except:
                    subdomain = subdomain + "1"
                domain = (
                    self._gis.url[:8]
                    + subdomain
                    + "-"
                    + self._gis.properties["urlKey"]
                    + ".hub.arcgis.com"
                )

        # For Enterprise Sites
        else:
            item_type = "Site Application"
            typekeywords = (
                "Hub, hubSite, hubSolution, hubsubdomain|"
                + subdomain
                + ", JavaScript, Map, Mapping Site, Online Map, OpenData, Ready To Use, selfConfigured, Web Map"
            )
            tags = ["Enterprise Site"]
            description = (
                "DO NOT DELETE OR MODIFY THIS ITEM. This item is managed by the ArcGIS Enterprise Sites application. To make changes to this site, please visit"
                + self._gis.url
                + "apps/sites/admin/"
            )

            # Domain manipulation
            domain = "https://" + self._gis.url[8:-5] + "apps/sites/#/" + subdomain

            # Check if site subdomain exists
            if self._gis.content.search(
                query="typekeywords:hubsubdomain|" + subdomain + " AND title:" + title
            ):
                print(
                    self._gis.content.search(
                        query="typekeywords:hubsubdomain|" + subdomain
                    )
                )
                raise ValueError(
                    "You already have a site that uses this subdomain. Please provide another title."
                )

        # setting item properties based on type of site
        if self._gis._portal.is_arcgisonline and self._hub._hub_enabled:
            # Hub Premium Site
            # if self._hub._hub_enabled:
            content_group_id = self.initiative.content_group_id
            collab_group_id = self.initiative.collab_group_id
            _item_dict = {
                "type": item_type,
                "typekeywords": typekeywords,
                "tags": tags,
                "title": title,
                "description": description,
                "culture": self._gis.properties.user.culture,
                "properties": {
                    "hasSeenGlobalNav": True,
                    "createdFrom": "defaultInitiativeSiteTemplate",
                    "schemaVersion": 1.5,
                    "contentGroupId": content_group_id,
                    "followersGroupId": self.initiative.followers_group_id,
                    "parentInitiativeId": self.initiative.itemid,
                    "children": [],
                },
                "url": domain,
            }
            if collab_group_id:
                collab_group = self._gis.groups.get(collab_group_id)
                _item_dict["properties"]["collaborationGroupId"] = collab_group_id
            _datafile = "init-sites-data.json"

        # Non Hub Premium Sites
        else:
            # Checking if it is a Hub Basic site
            if self._gis._portal.is_arcgisonline:
                _content_group_title = title + " Content"
                _content_group_dict = {
                    "title": _content_group_title,
                    "tags": ["Hub Group", "Hub Content Group", "Hub Site Group"],
                    "access": "public",
                }
                _collab_group_title = title + " Core Team"
                _collab_group_dict = {
                    "title": _collab_group_title,
                    "tags": [
                        "Hub Group",
                        "Hub Site Group",
                        "Hub Core Team Group",
                        "Hub Team Group",
                    ],
                    "access": "org",
                    "capabilities": "updateitemcontrol",
                    "membershipAccess": "collaboration",
                    "snippet": "Members of this group can create, edit, and manage the site, pages, and other content related to "
                    + title
                    + ".",
                }
                created_from = "basicDefaultSite Solution Template (embedded)"
            else:
                # Defining content, collaboration groups for Enterprise Sites
                collab_group_id = None
                _content_group_dict = {
                    "title": subdomain + " Content",
                    "tags": ["Sites Group", "Sites Content Group"],
                    "access": "org",
                    "snippet": "Applications, maps, data, etc. shared with this group generates the "
                    + subdomain
                    + " content catalog.",
                }
                _collab_group_dict = {
                    "title": subdomain + " Core Team",
                    "tags": ["Sites Group", "Sites Core Team Group"],
                    "access": "org",
                    "capabilities": "updateitemcontrol",
                    "membershipAccess": "org",
                    "snippet": "Members of this group can create, edit, and manage the site, pages, and other content related to "
                    + subdomain
                    + ".",
                }
                created_from = "portalDefaultSite"
            # Create groups
            content_group = self._gis.groups.create_from_dict(_content_group_dict)
            content_group_id = content_group.id
            # Create collaboration group if necessary privileges exist
            if self._gis.users.me.role == "org_admin":
                collab_group = self._gis.groups.create_from_dict(_collab_group_dict)
                collab_group.protected = True
                collab_group_id = collab_group.id
            else:
                collab_group_id = None
            # Protect groups from accidental deletion
            content_group.protected = True
            _item_dict = {
                "type": item_type,
                "typekeywords": typekeywords,
                "tags": tags,
                "title": title,
                "description": description,
                "properties": {
                    "hasSeenGlobalNav": True,
                    "createdFrom": created_from,
                    "schemaVersion": 1.5,
                    "contentGroupId": content_group_id,
                    "children": [],
                },
                "url": domain,
            }
            if collab_group_id is not None:
                _item_dict["properties"]["collaborationGroupId"] = collab_group.id
            _datafile = "sites-data.json"

        # Create site item, share with group
        site = self._gis.content.add(_item_dict, owner=self._gis.users.me.username)

        # Share with necessary group if group exists
        try:
            site.share(groups=[collab_group])
        except:
            pass

        # protect site from accidental deletion
        site.protect(enable=True)

        # Setting data for site
        data_path = os.path.abspath(
            os.path.join(os.path.dirname(__file__), "_store/" + _datafile)
        )

        with open(data_path) as f:
            _site_data = json.load(f)

        # Register site and update its data
        _data = self._create_and_register_site(
            site, subdomain, _site_data, content_group_id, collab_group_id
        )
        _data = json.dumps(_data)
        site.update(item_properties={"text": _data, "url": domain})
        return Site(self._gis, site)

    def clone(self, site, pages=True, title=None):
        """
        Clone allows for the creation of a site that is derived from the current site.

        .. note::
            Use this method if you are cloning a Site object from a Hub Basic or Enterprise environment.
            To clone from Hub Premium environments, please use the `initiatives.clone` method.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        site                Required Site object of site to be cloned.
        ---------------     --------------------------------------------------------------------
        pages               Optional Boolean. Decides if pages will be copied. Default is True.
        ---------------     --------------------------------------------------------------------
        title               Optional String.
        ===============     ====================================================================

        :return:
           Site.
        """
        collab_group_id = None
        from datetime import timezone

        now = datetime.now(timezone.utc)
        # Checking if item of correct type has been passed
        if "hubSite" not in site.item.typeKeywords:
            raise Exception("Incorrect item type. Site item needed for cloning.")
        # New title
        if title is None:
            title = site.title + "-copy-%s" % int(now.timestamp() * 1000)
        if self.initiative is None:
            try:
                if self._hub._hub_enabled:
                    return self._hub.initiatives.add(title, site=site)
            except:
                pass
        subdomain = title.replace(" ", "-").lower()
        # For Hub Sites
        if self._gis._portal.is_arcgisonline:
            item_type = "Hub Site Application"
            typekeywords = "Hub, hubSite, hubSolution, JavaScript, Map, Mapping Site, Online Map, OpenData, Ready To Use, selfConfigured, Web Map, Registered App"
            description = "DO NOT DELETE OR MODIFY THIS ITEM. This item is managed by the ArcGIS Hub application. To make changes to this site, please visit https://hub.arcgis.com/admin/"
            domain = (
                self._gis.url[:8]
                + subdomain
                + "-"
                + self._gis.properties["urlKey"]
                + ".hub.arcgis.com"
            )
        # For Enterprise Sites
        else:
            item_type = "Site Application"
            typekeywords = (
                "Hub, hubSite, hubSolution, hubsubdomain|"
                + subdomain
                + ", JavaScript, Map, Mapping Site, Online Map, OpenData, Ready To Use, selfConfigured, Web Map"
            )
            domain = "https://" + self._gis.url[8:-5] + "apps/sites/#/" + subdomain
        _site_properties = {
            "type": item_type,
            "typekeywords": typekeywords,
            "tags": ["Enterprise Site"],
            "title": title,
            "url": domain,
        }
        # Updating properties, groups for Hub sites
        if self.initiative is not None:
            content_group_id = self.initiative.content_group_id
            collab_group_id = self.initiative.collab_group_id
            _site_properties["properties"] = {
                "hasSeenGlobalNav": True,
                "createdFrom": "defaultInitiativeSiteTemplate",
                "schemaVersion": 1.5,
                "contentGroupId": content_group_id,
                "parentInitiativeId": self.initiative.itemid,
                "children": [],
            }
            if self._hub._hub_enabled:
                _site_properties["properties"][
                    "followersGroupId"
                ] = self.initiative.followers_group_id
            if collab_group_id:
                collab_group = self._gis.groups.get(collab_group_id)
                _site_properties["properties"]["collaborationGroupId"] = collab_group_id
        else:
            # Defining content, collaboration groups for Hub Basic and Enterprise Sites
            # For Hub Basic Sites
            if self._gis._portal.is_arcgisonline:
                _content_group_dict = {
                    "title": subdomain + " Content",
                    "tags": ["Hub Group", "Hub Content Group", "Hub Site Group"],
                    "access": "public",
                }
                _collab_group_dict = {
                    "title": subdomain + " Core Team",
                    "tags": [
                        "Hub Group",
                        "Hub Site Group",
                        "Hub Core Team Group",
                        "Hub Team Group",
                    ],
                    "access": "org",
                    "capabilities": "updateitemcontrol",
                    "membershipAccess": "collaboration",
                    "snippet": "Members of this group can create, edit, and manage the site, pages, and other content related to "
                    + subdomain
                    + ".",
                }
                created_from = "basicDefaultSite Solution Template (embedded)"
            # For Enterprise Sites
            else:
                _content_group_dict = {
                    "title": subdomain + " Content",
                    "tags": ["Sites Group", "Sites Content Group"],
                    "access": "org",
                    "snippet": "Applications, maps, data, etc. shared with this group generates the "
                    + subdomain
                    + " content catalog.",
                }
                _collab_group_dict = {
                    "title": subdomain + " Core Team",
                    "tags": ["Sites Group", "Sites Core Team Group"],
                    "access": "org",
                    "capabilities": "updateitemcontrol",
                    "membershipAccess": "org",
                    "snippet": "Members of this group can create, edit, and manage the site, pages, and other content related to "
                    + subdomain
                    + ".",
                }
                created_from = "portalDefaultSite"
            # Create content group
            content_group = self._gis.groups.create_from_dict(_content_group_dict)
            content_group_id = content_group.id
            # Protect groups from accidental deletion
            content_group.protected = True
            _site_properties["properties"] = {
                "hasSeenGlobalNav": True,
                "createdFrom": created_from,
                "schemaVersion": 1.5,
                "contentGroupId": content_group_id,
            }
            # Create collaboration group if privilege exists
            if self._gis.users.me.role == "org_admin":
                collab_group = self._gis.groups.create_from_dict(_collab_group_dict)
                collab_group_id = collab_group.id
                collab_group.protected = True
                _site_properties["properties"]["collaborationGroupId"] = collab_group_id

        # Create site item, share with group
        new_item = self._gis.content.add(
            _site_properties, owner=self._gis.users.me.username
        )

        # Share with necessary group
        try:
            new_item.share(groups=[collab_group])
        except:
            pass

        # Register new site and update its data
        _data = self._create_and_register_site(
            new_item, subdomain, site.definition, content_group_id, collab_group_id
        )

        new_item.update(item_properties={"text": _data, "url": domain})
        new_site = Site(self._gis, new_item)

        # Pages of the site
        site_pages = site.pages.search()
        # If pages exist
        if len(site_pages) > 0:
            # Check the value of param
            if pages:
                replaced = str(new_site.definition)
                for page in site_pages:
                    try:
                        new_site.pages.unlink(page)
                    except:
                        pass
                    new_page = new_site.pages.clone(page)
                    # Replacing references to this page with cloned page in site's layout
                    replaced = re.sub(page.itemid, new_page.itemid, replaced)
                    replaced = re.sub(page.title, new_page.title, replaced)
                    replaced = re.sub(page.slug, new_page.slug, replaced)
                new_data = json.loads(replaced)
                new_site.item.update(item_properties={"text": new_data})
        return new_site

    def get(self, site_id: str) -> Site:
        """Returns the site object for the specified site_id.

        =======================    =============================================================
        **Parameter**               **Description**
        -----------------------    -------------------------------------------------------------
        site_id                    Required string. The site itemid.
        =======================    =============================================================

        :return:
            The site object if the item is found, None if the item is not found.

        .. code-block:: python

            USAGE EXAMPLE: Fetch a site successfully

            site1 = myHub.sites.get('itemId12345')
            site1.item
        """
        siteItem = self._gis.content.get(site_id)
        if "hubSite" in siteItem.typeKeywords:
            return Site(self._gis, siteItem)
        else:
            raise TypeError("Item is not a valid site or is inaccessible.")

    def get_by_domain(self, domain_url: str) -> Site | None:
        """Returns the site object for the specified domain url.

        =======================    =============================================================
        **Parameter**               **Description**
        -----------------------    -------------------------------------------------------------
        domain_url                 Required string. The site url.
        =======================    =============================================================

        :return:
            The site object if the item is found, None if the item is not found.

        .. note::
            You may only fetch sites by domain local to your environment.
            E.g. If your Hub instance is an ArcGIS Online instance, then you can
            fetch ArcGIS Online sites by url, and if you have signed into an ArcGIS
            Enterprise Instance, only sites on premise will be available.

        .. code-block:: python

            USAGE EXAMPLE: Fetch a site successfully

            site1 = myHub.sites.get_by_domain('opendata.dc.gov')
            site1.item
        """
        # Check if Hub(GIS) is an ArcGIS Online instance
        if self._gis._portal.is_arcgisonline:
            if "http" in domain_url:
                domain_url = urlparse(domain_url).netloc
            path = "https://hub.arcgis.com/api/v3/domains/" + domain_url
            # fetch site itemid from domain service
            session = self._gis._con._session
            headers = {k: v for k, v in session.headers.items()}
            headers["Content-Type"] = "application/json"
            headers["Authorization"] = "X-Esri-Authorization"
            _site_domain = self._gis._con.get(path, headers=headers)
            try:
                siteId = _site_domain["siteId"]
            except KeyError:
                raise Exception(
                    "Domain record not found. Please check your domain_url."
                )
            return self.get(siteId)
        # For ArcGIS Enterprise
        else:
            sitelist = []
            subdomain = domain_url.split("#/", 1)[1]
            _query = "hubsubdomain|" + subdomain
            items = self._gis.content.search(
                query="typekeywords:hubSite," + _query, max_items=5000
            )
            # Return searched sites
            for item in items:
                sitelist.append(Site(self._gis, item))
            return sitelist

    def search(
        self,
        title: str | None = None,
        owner: str | None = None,
        created: str | None = None,
        modified: str | None = None,
        tags: str | None = None,
    ):
        """
        Searches for sites.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        title               Optional string. Return sites with provided string in title.
        ---------------     --------------------------------------------------------------------
        owner               Optional string. Return sites owned by a username.
        ---------------     --------------------------------------------------------------------
        created             Optional string. Date the site was created.
                            Shown in milliseconds since UNIX epoch.
        ---------------     --------------------------------------------------------------------
        modified            Optional string. Date the site was last modified.
                            Shown in milliseconds since UNIX epoch
        ---------------     --------------------------------------------------------------------
        tags                Optional string. User-defined tags that describe the site.
        ===============     ====================================================================

        :return:
           A list of matching sites.
        """

        sitelist = []

        if self.initiative is not None:
            _site_id = self.initiative.site_id
            return self.get(_site_id)

        # Build search query
        query = "typekeywords:hubSite"
        if title != None:
            query += " AND title:" + title
        if owner != None:
            query += " AND owner:" + owner
        if created != None:
            query += " AND created:" + created
        if modified != None:
            query += " AND modified:" + modified
        if tags != None:
            query += " AND tags:" + tags

        # Search
        items = self._gis.content.search(query=query, max_items=5000)

        # Return searched sites
        for item in items:
            sitelist.append(Site(self._gis, item))
        return sitelist



from pathlib import Path
import json
import random
from ._arcgis_model import ArcGISModel, _EmptyData
import traceback
from .._utils.env import raise_fastai_import_error
from ._codetemplate import image_classifier_prf

try:
    from ._change_detector_utils import get_learner
    from .._utils.change_detection_data import show_results
    from ._arcgis_model import _resnet_family
    from .._utils.common import _get_emd_path

    HAS_FASTAI = True
except ImportError:
    import_exception = traceback.format_exc()
    HAS_FASTAI = False


class ChangeDetector(ArcGISModel):

    """
    Creates a Change Detection model.

    A Spatial-Temporal Attention-Based Method and a New Dataset
    for Remote Sensing Image Change Detection -
    https://www.mdpi.com/2072-4292/12/10/1662

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    data                    Required fastai Databunch. Returned data object
                            from :meth:`~arcgis.learn.prepare_data`  function.
    ---------------------   -------------------------------------------
    backbone                Optional function. Backbone CNN model to be used
                            for creating the encoder of the :class:`~arcgis.learn.ConnectNet`,
                            which is `resnet18` by default. It supports
                            the ResNet family of backbones.
    ---------------------   -------------------------------------------
    attention_type          Optional string. It's value can be either be "PAM"
                            (Pyramid Attention Module) or "BAM"
                            (Basic Attention Module).
                            Defaults to "PAM".
    ---------------------   -------------------------------------------
    pretrained_path         Optional string. Path where pre-trained model is
                            saved.
    =====================   ===========================================

    :return: :class:`~arcgis.learn.ConnectNet` object
    """

    def __init__(
        self, data, backbone=None, attention_type="PAM", pretrained_path=None, **kwargs
    ):
        if not HAS_FASTAI:
            raise_fastai_import_error(
                import_exception=import_exception, message="", installation_steps=" "
            )

        if backbone is None:
            backbone = "resnet18"

        if not self._check_backbone_support(backbone):
            raise Exception(
                f"Enter only compatible backbones from {', '.join(self.supported_backbones)}"
            )

        super().__init__(data, backbone, pretrained_path=pretrained_path)
        backbone = self._backbone.__name__.lower()
        self.SA_type = attention_type
        self.learn = get_learner(self._data, backbone, self.SA_type)
        self._code = image_classifier_prf
        self._arcgis_init_callback()  # make first conv weights learnable
        if pretrained_path is not None:
            self.load(pretrained_path)

    @staticmethod
    def _available_metrics():
        return ["valid_loss", "precision", "recall", "f1"]

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return "<%s>" % (type(self).__name__)

    @property
    def supported_backbones(self):
        """Supported torchvision backbones for this model."""
        return ChangeDetector._supported_backbones()

    @staticmethod
    def _supported_backbones():
        return [*_resnet_family]

    @property
    def supported_datasets(self):
        """Supported dataset types for this model."""
        return ChangeDetector._supported_datasets()

    @staticmethod
    def _supported_datasets():
        return ["ChangeDetection", "Classified_Tiles"]

    @classmethod
    def from_model(cls, emd_path, data=None):
        """
        Creates a ChangeDetector model from an Esri Model Definition (EMD)
        file.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to Deep Learning Package
                                (DLPK) or Esri Model Definition(EMD) file.
        ---------------------   -------------------------------------------
        data                    Optional fastai Databunch. Returned
                                data object from :meth:`~arcgis.learn.prepare_data`  function or
                                None for inferencing.
        =====================   ===========================================

        :return: :class:`~arcgis.learn.ConnectNet` Object
        """
        emd_path = _get_emd_path(emd_path)
        emd_path = Path(emd_path)
        with open(emd_path) as f:
            emd = json.load(f)

        model_file = Path(emd["ModelFile"])
        if not model_file.is_absolute():
            model_file = emd_path.parent / model_file

        model_params = emd["ModelParameters"]
        if data is None:
            data = _EmptyData(
                path=emd_path.parent,
                loss_func=None,
                c=2,  # change, no_change
                chip_size=emd["ImageHeight"],
            )
            data._is_empty = True
            data._imagery_type = None
            data.emd_path = emd_path
            data.emd = emd
            for key, value in emd["DataAttributes"].items():
                setattr(data, key, value)

        return cls(data, **model_params, pretrained_path=str(model_file))

    @property
    def _model_metrics(self):
        metrics = self._get_model_metrics()
        return {
            "precision": float(metrics[0]),
            "recall": float(metrics[1]),
            "f1": float(metrics[2]),
        }

    def _get_model_metrics(self, **kwargs):
        checkpoint = getattr(self, "_is_checkpointed", False)
        if not hasattr(self.learn, "recorder"):
            return [0.0, 0.0, 0.0]

        if len(self.learn.recorder.metrics) == 0:
            return [0.0, 0.0, 0.0]

        model_accuracy = self.learn.recorder.metrics[-1]
        if checkpoint:
            val_losses = self.learn.recorder.val_losses
            model_accuracy = self.learn.recorder.metrics[
                self.learn._best_epoch  # index using best epoch.
            ]

        return model_accuracy

    def _get_emd_params(self, save_inference_file):
        _emd_template = {"DataAttributes": {}, "ModelParameters": {}}
        # add encoder parameters
        _emd_template["ModelParameters"]["attention_type"] = self.SA_type
        # chip size
        _emd_template["DataAttributes"]["chip_size"] = self._data.chip_size
        _emd_template["DataAttributes"][
            "_is_multispectral"
        ] = self._data._is_multispectral
        if self._data._is_multispectral:
            _emd_template["DataAttributes"]["_imagery_type"] = self._data._imagery_type
            _emd_template["DataAttributes"]["_bands"] = self._data._bands
            _emd_template["DataAttributes"]["_rgb_bands"] = self._data._rgb_bands
            _emd_template["DataAttributes"][
                "_extract_bands"
            ] = self._data._extract_bands
            _emd_template["DataAttributes"]["_train_tail"] = self._data._train_tail
            _emd_template["DataAttributes"][
                "_band_min_values"
            ] = self._data._band_min_values.tolist()
            _emd_template["DataAttributes"][
                "_band_max_values"
            ] = self._data._band_max_values.tolist()

        # normalization stats
        norm_stats = []
        for k in self._data.norm_stats:
            norm_stats.append(k)
        _emd_template["DataAttributes"]["norm_stats"] = list(norm_stats)
        _emd_template["DataAttributes"]["class_mapping"] = self._data.class_mapping
        _emd_template["DataAttributes"]["color_mapping"] = self._data.color_mapping
        _emd_template["DataAttributes"]["classes"] = self._data.classes
        _emd_template["DataAttributes"]["batch_size"] = self._data.batch_size
        _emd_template["Framework"] = "arcgis.learn.models._inferencing"
        # object classifier config can be used.
        _emd_template["ModelConfiguration"] = "change_detection"
        # Model Type
        _emd_template["ModelType"] = "ImageClassification"
        # Inference function of object classifier.
        if save_inference_file:
            _emd_template["InferenceFunction"] = "ArcGISImageClassifier.py"
        else:
            _emd_template[
                "InferenceFunction"
            ] = "[Functions]System\\DeepLearning\\ArcGISLearn\\ArcGISImageClassifier.py"
        if self._is_multispectral:
            # change this when we start to honour extract bands parameter.
            _emd_template["ExtractBands"] = list(
                range(len(self._data._extract_bands) * 2)
            )
        else:
            # To extract bands from concatenated images (RGB)
            _emd_template["ExtractBands"] = list(range(6))
        _emd_template["Classes"] = []
        class_data = {}
        for (
            value,
            class_name,
        ) in self._data.class_mapping.items():  # 0th index is background
            class_data["Value"] = value
            class_data["Name"] = class_name
            class_data["Color"] = self._data.color_mapping[value]
            _emd_template["Classes"].append(class_data.copy())

        return _emd_template

    def show_results(self, rows=4, **kwargs):
        """
        Displays the results of a trained model on the validation set.
        """
        show_results(self, rows, **kwargs)

    def precision_recall_score(self):
        """
        Computes precision, recall and f1 score.
        """
        from .._utils.classified_tiles import per_class_metrics

        return per_class_metrics(self, postprocess_type="CD")

    def _score_author(self):
        """
        Internal implementation required for comparing
        the author's way of computing scores.
        """
        from ._change_detector_utils import calculate_author_metrics

        return calculate_author_metrics(self)

    def predict(self, before_image, after_image, **kwargs):
        """
        Predict on a pair of images.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        before_image            Required string. Path to image from before.
        ---------------------   -------------------------------------------
        after_image             Required string. Path to image from later.
        =====================   ===========================================

        **Kwargs**

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        crop_predict            Optional Boolean. If True, It will predict
                                using a sliding window strategy. Typically, used
                                when image size is larger than the `chip_size`
                                the model is trained on. Default False.
        ---------------------   -------------------------------------------
        visualize               Optional Boolean. If True, It will plot
                                the predictions on the notebook. Default False.
        ---------------------   -------------------------------------------
        save                    Optional Boolean. If true will write the
                                prediction file on the disk. Default False.
        =====================   ===========================================

        :return: PyTorch Tensor of the change mask.
        """
        from .._utils.change_detection_data import predict

        return predict(self, before_image, after_image, **kwargs)



from arcgis import GIS
from typing import Optional, Dict, Union, List
from ._task import Task
from ._util import _Util


class RealTimeAnalytics(Task):
    """
    the ``RealTimeAnalytics`` class implements Task and provides public facing methods to
    access RealTimeAnalytics API endpoints.
    """

    _id = ""
    _gis = None
    _util = None
    _item = None

    def __init__(self, gis: GIS, util: _Util, item: Optional[Dict] = None):
        self._gis = gis
        self._util = util

        if item:
            self._item = item
            self._id = item["id"]

    # ----------------------------------------------------------------------
    def __repr__(self):
        return "<%s id:%s label:%s>" % (
            type(self).__name__,
            self._id,
            self._item["label"],
        )

    # ----------------------------------------------------------------------
    def start(self) -> Dict:
        """
        Start the Real-Time Analytics for the given ID.

        :return: response of realtime_analytics start

        .. code-block:: python

            # Start real-time analytics

            # Method: <item>.start()

            sample_realtime_task.start()

        """
        return self._util._start("analytics/realtime", self._id)

    # ----------------------------------------------------------------------
    def stop(self) -> Dict:
        """
        Stop the Real-Time Analytics for the given ID.
        Return True if the the Real-Time Analytics was successfully stopped.

        :return: boolean

        .. code-block:: python

            # Stop real-time analytics

            # Method: <item>.stop()

            sample_realtime_task.stop()

        """
        return self._util._stop("analytics/realtime", self._id)

    # ----------------------------------------------------------------------
    @property
    def status(self) -> Dict:
        """
        Get the status of the running Real-Time Analytics for the given ID.

        :return: response of Real-Time Analytics status

        .. code-block:: python

            # Retrieve status of real-time analytics task

            # Property: <item>.status()

            status = sample_realtime_task.status
            status

        """
        return self._util._status("analytics/realtime", self._id)

    # ----------------------------------------------------------------------
    @property
    def metrics(self) -> Dict:
        """
        Get the metrics of the running Real-Time Analytics for the given ID.

        :return: response of Real-Time Analytics metrics

        .. code-block:: python

            # Retrieve metrics of real-time analytics task

            # Property: <item>.metrics()

            metrics = sample_realtime_task.metrics
            metrics

        """
        return self._util._metrics("analytics/realtime/metrics", self._id)

    # ----------------------------------------------------------------------
    def delete(self) -> bool:
        """
        Deletes an existing Real-Time Analytics task instance.

        :return: A boolean containing True (for success) or False (for failure) a dictionary with details is returned.

        .. code-block:: python

            # Delete a real-time analytics

            # Method: <item>.delete()

            sample_realtime_task.delete

        """
        return self._util._delete("analytics/realtime", self._id)



"""
Modified from requests_toolbelt's GuesAuth to handle NTLM and Kerbos

"""
from requests import auth
from requests import cookies

try:
    from arcgis.auth import EsriWindowsAuth

    HAS_SSPI = True
except ImportError:
    HAS_SSPI = False

try:
    from arcgis.auth import EsriKerberosAuth

    HAS_KERBEROS = True
except ImportError:
    HAS_KERBEROS = False
from arcgis.auth import EsriBasicAuth
from arcgis.auth._auth._schain import SupportMultiAuth
from requests_toolbelt.auth import _digest_auth_compat as auth_compat, http_proxy_digest


class GuessAuth(auth.AuthBase, SupportMultiAuth):
    """Guesses the auth type by the WWW-Authentication header."""

    _try_auth_count = None

    def __init__(self, username, password, **kwargs):
        self.username = username
        self.password = password
        self.auth = None
        self.pos = None
        self._try_auth_count = 0
        self.proxies = kwargs.pop("proxies", {})
        self._legacy = kwargs.pop("legacy", True)

    def _handle_basic_auth_401(self, r, kwargs):
        if self.pos is not None:
            r.request.body.seek(self.pos)

        # Consume content and release the original connection
        # to allow our new request to reuse the same one.
        r.content
        r.raw.release_conn()
        prep = r.request.copy()
        if not hasattr(prep, "_cookies"):
            prep._cookies = cookies.RequestsCookieJar()
        cookies.extract_cookies_to_jar(prep._cookies, r.request, r.raw)
        prep.prepare_cookies(prep._cookies)

        self.auth = EsriBasicAuth(self.username, self.password, "http", False)
        prep = self.auth(prep)
        _r = r.connection.send(prep, **kwargs)
        _r.history.append(r)
        _r.request = prep

        return _r

    def _handle_ntlm_auth_401(self, r, kwargs):
        self.auth = EsriWindowsAuth(self.username, self.password, legacy=self._legacy)
        try:
            self.auth.init_per_thread_state()
        except AttributeError:
            # If we're not on requests 2.8.0+ this method does not exist and
            # is not relevant.
            pass

        # Check that the attr exists because much older versions of requests
        # set this attribute lazily. For example:
        # https://github.com/kennethreitz/requests/blob/33735480f77891754304e7f13e3cdf83aaaa76aa/requests/auth.py#L59
        if hasattr(self.auth, "num_401_calls") and self.auth.num_401_calls is None:
            self.auth.num_401_calls = 1
        # Digest auth would resend the request by itself. We can take a
        # shortcut here.
        return self.auth.response_hook(r, **kwargs)

    def _handle_kerb_auth_401(self, r, kwargs):
        self.auth = EsriKerberosAuth(proxies=self.proxies)
        try:
            self.auth.init_per_thread_state()
        except AttributeError:
            # If we're not on requests 2.8.0+ this method does not exist and
            # is not relevant.
            pass

        # Check that the attr exists because much older versions of requests
        # set this attribute lazily. For example:
        # https://github.com/kennethreitz/requests/blob/33735480f77891754304e7f13e3cdf83aaaa76aa/requests/auth.py#L59
        if hasattr(self.auth, "num_401_calls") and self.auth.num_401_calls is None:
            self.auth.num_401_calls = 1
        # Digest auth would resend the request by itself. We can take a
        # shortcut here.
        return self.auth.handle_401(r, **kwargs)

    def _handle_digest_auth_401(self, r, kwargs):
        self.auth = auth_compat.HTTPDigestAuth(self.username, self.password)
        try:
            self.auth.init_per_thread_state()
        except AttributeError:
            # If we're not on requests 2.8.0+ this method does not exist and
            # is not relevant.
            pass

        # Check that the attr exists because much older versions of requests
        # set this attribute lazily. For example:
        # https://github.com/kennethreitz/requests/blob/33735480f77891754304e7f13e3cdf83aaaa76aa/requests/auth.py#L59
        if hasattr(self.auth, "num_401_calls") and self.auth.num_401_calls is None:
            self.auth.num_401_calls = 1
        # Digest auth would resend the request by itself. We can take a
        # shortcut here.
        return self.auth.handle_401(r, **kwargs)

    def handle_401(self, r, **kwargs):
        """Resends a request with auth headers, if needed."""

        www_authenticate = r.headers.get("www-authenticate", "").lower()

        if "basic" in www_authenticate:
            return self._handle_basic_auth_401(r, kwargs)

        if "digest" in www_authenticate:
            return self._handle_digest_auth_401(r, kwargs)
        ##########################################
        ##                                       #
        ##  Handles the NTLM/Kerberos Handshake  #
        ##                                       #
        ##########################################
        if www_authenticate.find("ntlm") > -1 and self.username and self.password:
            if self._try_auth_count == 0:
                self._try_auth_count += 1
                self.auth = EsriWindowsAuth(
                    self.username,
                    self.password,
                    verify_cert=False,
                    legacy=self._legacy,
                    proxies=self.proxies,
                )
                return self._handle_ntlm_auth_401(r, kwargs)
            elif self._try_auth_count == 1 and HAS_KERBEROS:
                self.auth = EsriKerberosAuth(self.proxies)
                self._try_auth_count += 1
                return self._handle_kerb_auth_401(r, kwargs)
            else:
                raise Exception("Could not login to the site.")
        elif www_authenticate.find("ntlm") > -1:
            if HAS_SSPI:
                self.auth = EsriWindowsAuth(
                    username=self.username,
                    password=self.password,
                    verify_cert=False,
                    legacy=self._legacy,
                    proxies=self.proxies,
                )
            else:
                self.auth = EsriWindowsAuth(
                    username=self.username,
                    password=self.password,
                    verify_cert=False,
                    legacy=self._legacy,
                    proxies=self.proxies,
                )

    def __call__(self, request):
        if self.auth is not None:
            return self.auth(request)

        try:
            self.pos = request.body.tell()
        except AttributeError:
            pass

        request.register_hook("response", self.handle_401)
        return request



"""set of common utilities"""
import os
import sys
import time
import uuid
import zipfile
import datetime
from datetime import date
import tempfile
from contextlib import contextmanager
import logging
import decimal
import functools


@functools.lru_cache(maxsize=35)
def _validate_url(url: str, gis: "GIS", url_type: str = None) -> str:
    """calculates the service URL"""
    finders = [
        "naserver",
        "gpserver",
        "mapserver",
        "featureserver",
        "imageserver",
        "geoenrichmentserver",
    ]
    part = None
    if not any([url.lower().endswith(f) for f in finders]):
        part = os.path.basename(url)
        url = os.path.dirname(url)
    try:
        res = gis._private_service_url(url)
    except:
        if part:
            return f"{url}/{part}"
        return url
    return_url = None
    if gis._is_hosted_nb_home and "privateServiceUrl" in res:
        return_url = res["privateServiceUrl"]
    else:
        if url_type is None or str(url_type).lower() == "public":
            if "serviceUrl" in res:
                return_url = res["serviceUrl"]
            else:
                return_url = url
        else:
            if "privateServiceUrl" in res:
                return_url = res["privateServiceUrl"]
            elif "serviceUrl" in res:
                return_url = res["serviceUrl"]
            else:
                return_url = url
    if part:
        return f"{return_url}/{part}"
    return return_url


# ----------------------------------------------------------------------
def bytesto(size, to="m", bsize=1024):
    """convert bytes to megabytes, etc.
    sample code:
        print('mb= ' + str(bytesto(314575262000000, 'm')))

    sample output:
        mb= 300002347.946
    """

    a = {"k": 1, "m": 2, "g": 3, "t": 4, "p": 5, "e": 6}
    r = float(size)
    for i in range(a[to]):
        r = r / bsize
    return r


# ----------------------------------------------------------------------
def create_uid():
    return uuid.uuid4().hex


# ----------------------------------------------------------------------
def inspect_function_inputs(fn, **params):
    """
    Given any function and a set of key/value pairs, where the ```params``` is a dictionary,
    the code inspects the input method and then returns a list of accepted inputs as
    a new dictionary.  This method is used primarily to validate GP services and ensure
    that the parameters given are supported in the current version of the tool.

    :return: dictionary

    Example:

    >>> def add(x,y):
    >>>    return x+y

    >>> valid_inputs = inspect_function_inputs(fn=add, **{'x' : 1, 'y': 2, 'cat' : 3})
    >>> print(valid_inputs)
    {'x' : 1, 'y': 2}

    """
    import inspect

    try:
        args = list(inspect.signature(fn).parameters.keys()) + ["estimate"]
    except ValueError:
        args = inspect.getfullargspec(func=fn).args
    if "gis" in args:
        args.pop(args.index("gis"))
    valid = {}
    for key in params.keys():
        if key in args and params[key] is not None:
            valid[key] = params[key]

    return valid


# ----------------------------------------------------------------------
def _date_handler(obj):
    import numpy

    npversion = [int(i) for i in numpy.__version__.split(".")]
    if npversion < [1, 20, 0]:
        FLOAT_CHECKER = (numpy.float, numpy.float32, numpy.float64)
    else:
        FLOAT_CHECKER = (float, numpy.float32, numpy.float64)
    from ._mixins import PropertyMap

    if type(obj) is datetime.date:
        import datetime as _dt

        obj = _dt.datetime.combine(obj, _dt.datetime.min.time())
    if isinstance(obj, datetime.datetime) or isinstance(obj, date):
        try:
            return local_time_to_online(obj)
        except:
            diff = datetime.datetime(1970, 1, 1) - obj
            v = -diff.total_seconds() * 1000
            return int(v)
    elif isinstance(obj, (numpy.int32, numpy.int64)):
        return _date_handler(int(obj))
    elif isinstance(obj, decimal.Decimal):
        return float(obj)
    elif isinstance(obj, FLOAT_CHECKER):
        return float(obj)
    elif isinstance(obj, numpy.ndarray):
        return obj.tolist()
    elif isinstance(obj, PropertyMap):
        return dict(obj)
    else:
        return obj


# ----------------------------------------------------------------------
def local_time_to_online(dt=None):
    """
    converts datetime object to a UTC timestamp for AGOL
    Inputs:
       dt - datetime object
    Output:
       Long value
    """

    if dt is None:
        dt = datetime.datetime.now()

    if sys.version_info.major == 3:
        return int(dt.timestamp() * 1000)
    elif isinstance(dt, datetime.datetime) and dt.tzinfo:
        dt = dt.astimezone()

    return int(time.mktime(dt.timetuple()) * 1000)


# ----------------------------------------------------------------------
def online_time_to_string(value, timeFormat):
    """
    Converts a timestamp to date/time string
    Inputs:
       value - timestamp as long
       timeFormat - output date/time format
    Output:
       string
    """
    return datetime.datetime.fromtimestamp(value / 1000).strftime(timeFormat)


# ----------------------------------------------------------------------
def timestamp_to_datetime(timestamp):
    """
    Converts a timestamp to a datetime object
    Inputs:
       timestamp - timestamp value as Long
    output:
       datetime object
    """
    return datetime.datetime.fromtimestamp(timestamp / 1000)


###########################################################################
class Error(Exception):
    pass


# --------------------------------------------------------------------------
@contextmanager
def _tempinput(data):
    temp = tempfile.NamedTemporaryFile(delete=False)
    temp.write((bytes(data, "UTF-8")))
    temp.close()
    yield temp.name
    os.unlink(temp.name)


# --------------------------------------------------------------------------
def _lazy_property(fn):
    """Decorator that makes a property lazy-evaluated."""
    # http://stevenloria.com/lazy-evaluated-properties-in-python/
    attr_name = "_lazy_" + fn.__name__

    @property
    @functools.wraps(fn)
    def _lazy_property(self):
        if not hasattr(self, attr_name):
            setattr(self, attr_name, fn(self))
        return getattr(self, attr_name)

    return _lazy_property


# --------------------------------------------------------------------------
def _is_shapefile(data):
    if zipfile.is_zipfile(data):
        zf = zipfile.ZipFile(data, "r")
        namelist = zf.namelist()
        for name in namelist:
            if name.endswith(".shp") or name.endswith(".SHP"):
                return True
    return False


# --------------------------------------------------------------------------
def rot13(s):
    result = ""

    # Loop over characters.
    for v in s:
        # Convert to number with ord.
        c = ord(v)

        # Shift number back or forward.
        if c >= ord("a") and c <= ord("z"):
            if c > ord("m"):
                c -= 13
            else:
                c += 13
        elif c >= ord("A") and c <= ord("Z"):
            if c > ord("M"):
                c -= 13
            else:
                c += 13

        # Append to result.
        result += chr(c)

    # Return transformation.
    return result


# --------------------------------------------------------------------------
def zipws(path, outfile, keep=True):
    """
    compress the contents of a folder
    Parameters:
     :path: folder or folder contents to compress as a zip file
     :outfile: output file and location
     :keep: boolean - if true, the folder structure is kept, else just the
      files
    Output:
     path to a compressed zip file.
    """
    zipobj = zipfile.ZipFile(outfile, "w", zipfile.ZIP_DEFLATED)
    path = os.path.normpath(path)
    for dirpath, dirnames, filenames in os.walk(path):
        for file in filenames:
            if not file.endswith(".lock") and not file.endswith(".zip"):
                try:
                    if keep:
                        zipobj.write(
                            os.path.join(dirpath, file),
                            os.path.join(
                                os.path.basename(path),
                                os.path.join(dirpath, file)[len(path) + len(os.sep) :],
                            ),
                        )
                    else:
                        zipobj.write(
                            os.path.join(dirpath, file),
                            os.path.join(dirpath[len(path) :], file),
                        )
                except Exception:
                    pass
    zipobj.close()
    return outfile


# --------------------------------------------------------------------------
def _to_utf8(data):
    """Converts strings and collections of strings from unicode to utf-8."""
    if isinstance(data, dict):
        return {
            _to_utf8(key): _to_utf8(value)
            for key, value in data.items()
            if value is not None
        }
    elif isinstance(data, list):
        return [_to_utf8(element) for element in data]
    elif isinstance(data, str):
        return data
    elif isinstance(data, str):
        return data.encode("utf-8")
    elif isinstance(data, (float, int)):
        return data
    else:
        return data


# --------------------------------------------------------------------------
class _DisableLogger:
    def __enter__(self):
        logging.disable(logging.CRITICAL)

    def __exit__(self, a, b, c):
        logging.disable(logging.NOTSET)


# --------------------------------------------------------------------------
def chunks(l, n):
    """yield successive n-sized chunks from l."""
    for i in range(0, len(l), n):
        yield l[i : i + n]


# --------------------------------------------------------------------------
def is_pdf_file(file_path):
    """check the file first bytes to match with pdf signature"""
    with open(file_path, "rb") as f:
        return f.read(4) == b"%PDF"



#!/usr/bin/env python3

"""
The methods are taken from => https://github.com/anilbatra2185/road_connectivity

MIT License
==========================================================================

Copyright (c) 2020 Anil Batra

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""


import math

import numpy as np
from skimage.morphology import skeletonize

from .graph_utils import segmets_to_linestrings, simplify_graph, unique
from .sknw import build_sknw


def getKeypoints(mask, is_gaussian=True, thresh=0.8, is_skeleton=False, smooth_dist=4):
    """
    Generate keypoints for binary prediction mask.

    @param mask: Binary road probability mask
    @param gaussian: Flag to check if the given mask is gaussian/probability mask
                    from prediction
    @param thresh: Probability threshold used to cnvert the mask to binary 0/1 mask.
                   It is used to convert the gaussian mask with required road width
    @param is_skeleton: Flag to perform opencv skeletonization on the binarized
                        road mask
    @param smooth_dist: Tolerance parameter used to smooth the graph using
                        RDP algorithm

    @return: return ndarray of road keypoints
    """
    if np.max(np.unique(mask)) > 1:
        mask /= 255.0

    if is_gaussian:
        mask[mask < thresh] = 0
        mask[mask >= thresh] = 1

    h, w = mask.shape
    if is_skeleton:
        ske = mask
    else:
        ske = skeletonize(mask).astype(np.uint16)
    graph = build_sknw(ske, multi=True)

    segments = simplify_graph(graph, smooth_dist)
    linestrings_1 = segmets_to_linestrings(segments)
    linestrings = unique(linestrings_1)

    keypoints = []
    for line in linestrings:
        linestring = line.rstrip("\n").split("LINESTRING ")[-1]
        points_str = linestring.lstrip("(").rstrip(")").split(", ")
        # If there is no road present
        if "EMPTY" in points_str:
            return keypoints
        points = []
        for pt_st in points_str:
            x, y = pt_st.split(" ")
            x, y = float(x), float(y)
            points.append([x, y])

            x1, y1 = points[0]
            x2, y2 = points[-1]
            zero_dist1 = math.sqrt((x1) ** 2 + (y1) ** 2)
            zero_dist2 = math.sqrt((x2) ** 2 + (y2) ** 2)

            if zero_dist2 > zero_dist1:
                keypoints.append(points[::-1])
            else:
                keypoints.append(points)
    return keypoints


def getVectorMapsAngles(shape, keypoints, theta=5, bin_size=10):
    """
    Convert Road keypoints obtained from road mask to orientation angle mask.
    Reference: Section 3.1
        https://anilbatra2185.github.io/papers/RoadConnectivityCVPR2019.pdf

    @param shape: Road Label/PIL image shape i.e. H x W
    @param keypoints: road keypoints generated from Road mask using
                        function getKeypoints()
    @param theta: thickness width for orientation vectors, it is similar to
                    thicknes of road width with which mask is generated.
    @param bin_size: Bin size to quantize the Orientation angles.

    @return: Retun ndarray of shape H x W, containing orientation angles per pixel.
    """

    im_h, im_w = shape
    vecmap = np.zeros((im_h, im_w, 2), dtype=np.float32)
    vecmap_angles = np.zeros((im_h, im_w), dtype=np.float32)
    vecmap_angles.fill(360)
    height, width, channel = vecmap.shape
    for j in range(len(keypoints)):
        for i in range(1, len(keypoints[j])):
            a = keypoints[j][i - 1]
            b = keypoints[j][i]
            ax, ay = a[0], a[1]
            bx, by = b[0], b[1]
            bax = bx - ax
            bay = by - ay
            norm = math.sqrt(1.0 * bax * bax + bay * bay) + 1e-9
            bax /= norm
            bay /= norm

            min_w = max(int(round(min(ax, bx) - theta)), 0)
            max_w = min(int(round(max(ax, bx) + theta)), width)
            min_h = max(int(round(min(ay, by) - theta)), 0)
            max_h = min(int(round(max(ay, by) + theta)), height)

            for h in range(min_h, max_h):
                for w in range(min_w, max_w):
                    px = w - ax
                    py = h - ay
                    dis = abs(bax * py - bay * px)
                    if dis <= theta:
                        vecmap[h, w, 0] = bax
                        vecmap[h, w, 1] = bay
                        _theta = math.degrees(math.atan2(bay, bax))
                        vecmap_angles[h, w] = (_theta + 360) % 360

    vecmap_angles = (vecmap_angles / bin_size).astype(int)
    return vecmap, vecmap_angles


def convertAngles2VecMap(shape, vecmapAngles, bin_size=10.0):
    """
    Helper method to convert Orientation angles mask to Orientation vectors.

    @params shape: Road mask shape i.e. H x W
    @params vecmapAngles: Orientation agles mask of shape H x W
    @param bin_size: Bin size to quantize the Orientation angles.

    @return: ndarray of shape H x W x 2, containing x and y values of vector
    """

    h, w = shape
    vecmap = np.zeros((h, w, 2), dtype=float)
    max_angle_bin = 360.0 / bin_size

    for h1 in range(h):
        for w1 in range(w):
            angle = vecmapAngles[h1, w1]
            if angle < max_angle_bin:
                angle *= bin_size
                if angle >= 180.0:
                    angle -= 360.0
                vecmap[h1, w1, 0] = math.cos(math.radians(angle))
                vecmap[h1, w1, 1] = math.sin(math.radians(angle))

    return vecmap


def convertVecMap2Angles(shape, vecmap, bin_size=10):
    """
    Helper method to convert Orientation vectors to Orientation angles.

    @params shape: Road mask shape i.e. H x W
    @params vecmap: Orientation vectors of shape H x W x 2

    @return: ndarray of shape H x W, containing orientation angles per pixel.
    """

    im_h, im_w = shape
    angles = np.zeros((im_h, im_w), dtype=float)
    angles.fill(360)

    for h in range(im_h):
        for w in range(im_w):
            x = vecmap[h, w, 0]
            y = vecmap[h, w, 1]
            angles[h, w] = (math.degrees(math.atan2(y, x)) + 360) % 360

    angles = (angles / bin_size).astype(int)
    return angles



""" Defines store functions for working with AssignmentTypes in Version 2 projects.
"""

from ... import workforce
from .utils import validate, add_features, update_features, remove_features


def get_assignment_type_v2(project, code=None, name=None):
    """Gets the identified AssignmentType. Exactly one form of identification should be provided.
    :param project:
    :param code: The AssignmentType GlobalID.
    :param name: The AssignmentType name.
    """
    assignment_types = get_assignment_types_v2(project)
    if code:
        return next((at for at in assignment_types if at.code == code), None)
    elif name:
        return next((at for at in assignment_types if at.name == name), None)
    return None


def get_assignment_types_v2(project):
    """Gets all AssignmentTypes for the project.
    :param project:
    :returns: A list of AssignmentTypes.
    """
    return query_assignment_types(project, "1=1")


def query_assignment_types(project, where="1=1"):
    """Executes a query against the assignment types table.
    :param project: The project in which to query assignment types.
    :param where: An ArcGIS where clause.
    :returns: list of Assignment Types
    """
    assignment_type_features = project.assignment_types_table.query(
        where, return_all_records=True
    ).features
    return [
        workforce.AssignmentType(project, feature)
        for feature in assignment_type_features
    ]


def add_assignment_type_v2(project, name):
    """
    Adds a new assignment type
    """
    assignment_type = workforce.AssignmentType(project, name=name)
    return add_assignment_types_v2(project, [assignment_type])[0]


def add_assignment_types_v2(project, assignment_types):
    """Adds an AssignmentType to a project.
    :param project:
    :param assignment_types: list of AssignmentTypes
    :raises ValidationError: Indicates that one or more assignment types failed validation.
    """
    for assignment_type in assignment_types:
        validate(assignment_type._validate)

    use_global_ids = False
    features = [assignment_type.feature for assignment_type in assignment_types]
    add_features(project.assignment_types_table, features, use_global_ids)
    project._update_cached_assignment_types()
    return assignment_types


def update_assignment_types_v2(project, assignment_types):
    """Updates the AssignmentTypes.
    :param project:
    :param assignment_types: list of AssignmentTypes
    :raises ValidationError: Indicates that one or more assignment types failed validation.
    """
    for assignment_type in assignment_types:
        validate(assignment_type._validate_for_update)
    features = [assignment_type.feature for assignment_type in assignment_types]
    update_features(project.assignment_types_table, features)
    project._update_cached_assignment_types()
    return assignment_types


def update_assignment_type_v2(project, assignment_type, name=None):
    """Updates an assignment type's name"""
    if name:
        assignment_type.name = name
    return update_assignment_types_v2(project, [assignment_type])[0]


def delete_assignment_types_v2(project, assignment_types):
    """Removes AssignmentTypes from the project.
    :param project:
    :param assignment_types: list of AssignmentTypes.
    """
    for assignment_type in assignment_types:
        validate(assignment_type._validate_for_remove)
    features = [assignment_type.feature for assignment_type in assignment_types]
    remove_features(project.assignment_types_table, features)
    project._update_cached_assignment_types()
